/**
	Copyright (c) 2009-2013 by Zoltan Palmer

	This file is part of FroZenLight.

    FroZenLight is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    FroZenLight is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with FroZenLight.  If not, see http://www.gnu.org/licenses/
*/
package calc;
import java.util.ArrayList;
import java.util.HashMap;

import javax.swing.JComboBox;
import javax.swing.SwingUtilities;
import javax.swing.event.TableModelEvent;
import javax.swing.event.TableModelListener;
import javax.swing.table.TableModel;

import math.IFunction;
import math.VecMath;
import parser.ExpressionTree;
import parser.ExpressionTreeBuilder;
import parser.Symbols;

public class CalculatorModel<T extends Number> implements TableModel {

	public static final String MODE_EVAL = "  =  ";
	public static final String MODE_ROOT = "  = 0";
	public static final String[] MODES = new String[]{MODE_EVAL,MODE_ROOT};
	private final String ROOT_VAR = "x";
	private final VecMath<T> math;
	private final ArrayList<TableModelListener> listeners;
	private final String[] columnNames;
	private final ArrayList<Row> rows;
	public CalculatorModel(String[] columnNames,VecMath<T> mt) {
		this.listeners = new ArrayList<TableModelListener>();
		this.columnNames = columnNames;
		this.math = mt;
		this.rows = new ArrayList<Row>();
		for(int i = 0;i<20;i++)
			rows.add(new Row());
	}
	public void updateTable() {
		SwingUtilities.invokeLater(new Thread() {

			public void run() {
				for(TableModelListener l:listeners)
					l.tableChanged(new TableModelEvent(CalculatorModel.this));
			}
		});
	}
	public int getRowCount() {
		return rows.size();
	}
	public int getColumnCount() {
		return columnNames.length;
	}
	public String getColumnName(int columnIndex) {
		if(columnIndex<columnNames.length)
			return columnNames[columnIndex];
		return null;
	}
	public Class<?> getColumnClass(int cIdx) {
		switch(cIdx) {
			case 0:
				return String.class;
			case 1:
				return String.class;
			case 2:
				return String.class;
			case 3:
				return JComboBox.class;
			case 4:
				return String.class;
		}
		return null;
	}
	public boolean isCellEditable(int rIdx,int cIdx) {
		switch(cIdx) {
			case 0:
				return false;
			case 1:
				return true;
			case 2:
				return true;
			case 3:
				return true;
			case 4:
				return true;
		}
		return false;
	}
	public Object getValueAt(int rIdx,int cIdx) {
		switch(cIdx) {
			case 0:
				return (rIdx+1)+"";
			case 1:
				return rows.get(rIdx).exprs;
			case 2:
				return rows.get(rIdx).params;
			case 3:
				return rows.get(rIdx).mode?MODE_ROOT:MODE_EVAL;
			case 4:
				T retVal = rows.get(rIdx).result;
				if(retVal!=null)
					if(math.util.precision(retVal)>math.PRECISION)
						retVal = math.util.copy(retVal,math.PRECISION);
				return math.util.toPlainString(retVal);
		}
		return null;
	}
	public void setValueAt(Object val,int rIdx,int cIdx) {
		switch(cIdx) {
			case 1:
				rows.get(rIdx).exprs = (String)val;
				break;
			case 2:
				rows.get(rIdx).params = (String)val;
				break;
			case 3:
				rows.get(rIdx).mode = ((String)val).compareTo(MODE_ROOT)==0;
		}
	}
	public void addTableModelListener(TableModelListener l) {
		listeners.add(l);
	}
	public void removeTableModelListener(TableModelListener l) {
		listeners.remove(l);
	}
	public void calculate() throws Exception {
		for(Row r:rows)
			r.result = null;
		updateTable();
		ExpressionTreeBuilder<T> builder = new ExpressionTreeBuilder<T>(math);
		ArrayList<T> vals = new ArrayList<T>();
		for(int i = 0;i<rows.size();i++) {
			Row r = rows.get(i);
			if(r.exprs==null||r.exprs.trim().compareTo("")==0) {
				vals.add(null);
				continue;
			}
			try {
				final ExpressionTree<T> exprTree = builder.build(r.exprs);
				T[] vs = vals.toArray(math.util.array(vals.size()));
				final HashMap<String,T> map = exprTree.buildValueMap(r.params,vs);
				if(!r.mode)
					r.result = exprTree.calculate(map);
				else {
					if(!exprTree.containsVariable(ROOT_VAR)) {
						String msg = "Can't calculate root: Variable "+ROOT_VAR+" is missing";
						throw new Exception(msg);
					}
					if(!map.containsKey(ROOT_VAR)) {
						String msg = "Can't calculate root: Initial value "+"for variable "
						        +ROOT_VAR+" is missing.";
						throw new Exception(msg);
					}
					IFunction<T> f = new IFunction<T>() {

						@Override
						public T f(T arg) throws Exception {
							map.put(ROOT_VAR,arg);
							return exprTree.calculate(map);
						}
					};
					r.result = math.findRoot(f,map.get(ROOT_VAR));
				}
				vals.add(r.result);
				updateTable();
			} catch(Exception ex) {
				int line = rows.indexOf(r)+1;
				String msg = "Error in line "+line+".\n"+ex.getMessage();
				throw new Exception(msg);
			}
		}
		updateTable();
	}
	private class Row {

		private String exprs;
		private String params;
		private boolean mode;
		private T result;
		public Row() {}
	}
	public boolean isCleared() {
		for(Row r:rows)
			if(r.result!=null)
				return false;
		return true;
	}
	public void clearResults() {
		for(Row r:rows)
			r.result = null;
		updateTable();
	}
	public void removeRow(int row) {
		if(rows.size()>row) {
			rows.remove(row);
			if(!updateArgs(row,false))
				clearResults();
			updateTable();
		}
	}
	public void insertRow(int row) {
		if(rows.size()>=row) {
			rows.add(row,new Row());
			updateArgs(row,true);
			updateTable();
		}
	}
	private boolean updateArgs(int row,boolean decOrInc) {
		boolean refOk = true;
		for(int k = row;k<rows.size();k++) {
			Row r = rows.get(k);
			if(r.params!=null) {
				StringBuffer sb = new StringBuffer(r.params);
				int i1 = sb.indexOf(Symbols.REFERENCE);
				while(i1!=-1) {
					int i2 = sb.indexOf(Symbols.DELIMITER,i1);
					if(i2==-1)
						i2 = sb.length();
					short line = Short.parseShort(sb.substring(i1+1,i2).trim());
					if(decOrInc) {
						if(line-1>=row)
							sb.replace(i1+1,i2,""+(line+1));
					} else {
						if(line-1==row) {
							sb.replace(i1+1,i2,"");
							refOk = false;
						} else if(line-1>row)
							sb.replace(i1+1,i2,""+(line-1));
					}
					i1 = sb.indexOf(Symbols.REFERENCE,i2);
				}
				r.params = sb.toString();
			}
		}
		return refOk;
	}
}
