/**
	Copyright (c) 2009-2013 by Zoltan Palmer

	This file is part of FroZenLight.

    FroZenLight is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    FroZenLight is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with FroZenLight.  If not, see http://www.gnu.org/licenses/
*/
package mvc;

import java.util.ArrayList;
import java.util.List;

import lattice.Box2D;
import lattice.IBox;
import lattice.Ray;
import lattice.RayPolygon;
import lattice.ReflectionEngine;
import lattice.ShapeContainer;
import math.VecMath;
import mvc.scan.Observables;
import mvc.scan.ScanListener;

import common.FloatConverter;
import common.Globals;
import common.IConverter;
import common.Log;

public final class ScanSession<T extends Number> {
	private T[] sincos;
	private PixelMap<T> pixMap;
	private T yStart;
	private int col;
	private T dy;
	private int maxPos;
	private int pos = 0;
	private int processed = 0;
	private boolean active = false;
	private Model<T> model;
	boolean stop = false;
	ShapeContainer<T> tCtr;
	String msg;
	public Observables obs;
	
	public ScanSession(Model<T> model){ this.model = model; }
	void init(PixelMap<T> pmap,Observables o) {
		model.initScanSession(pmap,this);
		this.obs = o;
		this.pixMap = pmap;
		T dphi = pmap.xLengthPix2World((short)1);
		Log.getIstc().logln("Calculating sin/cos for incremental rotation dphi...");
		this.sincos = model.getMath().sincos(dphi);
	}
	public IBox<T> getInitBox() {
		return pixMap.getWorldBox();
	}
	public void scanObservables() {
		VecMath<T> tmath = model.getMath();
		T[] dir = tmath.util.array(2);
		dir[0] = tmath.util.cast(-1);
		dir[1] = tmath.ZERO;
		T[] start = tmath.util.array(2);
		start[0] = tmath.neg(model.getWorldBox().getCorner()[0]);
		List<ScanListener> scanListeners = model.getScanListeners();
		synchronized(scanListeners) {
			for(ScanListener sl:scanListeners)
				sl.start();
		}
		List<ScanRbl> rbls = new ArrayList<ScanRbl>();
		for(int i = 0;i<Runtime.getRuntime().availableProcessors();i++){
			ReflectionEngine<T> reflCalc = 	new ReflectionEngine<T>(tmath);
			reflCalc.setOutputType(ReflectionEngine.OUTPUT_LOCAL);
			ShapeContainer<T> ctr = new ShapeContainer<T>(tCtr,tmath);
			RayPolygon<T> poly = new RayPolygon<T>(tmath.util,Globals.CALC_ACCURACY);
			Ray<T> scanRay = new Ray<T>(start.clone(),dir.clone(),tmath.util);
			IBox<Float> fbox;
			VecMath<Float> fmath = model.getFloatMath();
			fbox = new Box2D<Float>(fmath.ZERO2D.clone(),
					1f*Model.GRID_SIDE_LENGTH,fmath);
			rbls.add(new ScanRbl(reflCalc,ctr,poly,scanRay,fbox));
		}
		maxPos = obs.reflCnt[0].length-1;
		dy = pixMap.yLengthPix2World((short)1);
		yStart = pixMap.yPix2World((short)0);
		for(int i = 0;i<rbls.size();i++){
			if(i>0)
				new Thread(rbls.get(i)).start();
		}
		for(short i = 0;i<obs.reflCnt.length&&!stop;i++) {
			col = i;
			reset();
			for(int p = nextPos();p!=-1&&!stop;p = nextPos())
				rbls.get(0).scan(p);
			synchronized(obs) {
				while(processed<maxPos&&!stop)
					try {
						obs.wait();
					} catch (InterruptedException e) {}
			}
			for(int k = 0;k<rbls.size();k++)			
				model.rotateGrid(sincos,rbls.get(k).ctr);
		}
		stop = true;
		reset();
		synchronized(scanListeners) {
			for(ScanListener sl:scanListeners)
				sl.stop(msg);
		}
	}
	synchronized int nextPosOrWait(){
		while(!active)
			try{
				wait();
			} catch(InterruptedException ie){}
		if(pos>maxPos){
			try{
				wait();
			} catch(InterruptedException ie){}
			return nextPosOrWait();
		}
		return ++pos-1;
	}
	synchronized int nextPos() {return pos>maxPos?-1:++pos-1;}
	synchronized void reset(){
		pos = 0;
		active = true;
		processed = 0;
		notifyAll();
	}
	private class ScanRbl implements Runnable {
		private ReflectionEngine<T> reflCalc;
		private ShapeContainer<T> ctr;
		private RayPolygon<T> poly;
		private Ray<T> scanRay;
		private IBox<Float> fbox;
		ScanRbl(ReflectionEngine<T> re,ShapeContainer<T> sc,RayPolygon<T> rp,
				Ray<T> r,IBox<Float> fb){
			this.reflCalc = re;
			this.ctr = sc;
			this.poly = rp;
			this.scanRay = r;
			this.fbox = fb;
		}
		@Override
		public void run() {
			while(!stop) 
				scan(nextPosOrWait());
		}
		private void scan(int row){
			T[] start = scanRay.getStart();
			VecMath<T> tmath = model.getMath();
			start[1] = tmath.sub(yStart,tmath.mul(dy,row));
			reflCalc.calcDynamic(scanRay,ctr,poly,null);
			obs.reflCnt[col][row] = poly.size();
			VecMath<Float> fmath = model.getFloatMath();
			IConverter<T, Float> tfltCvtr = new FloatConverter<T>();
			obs.length[col][row] = poly.calcLength(fbox,tfltCvtr,fmath);
			obs.obs1[col][row] = obs.reflCnt[col][row]/obs.length[col][row];
			List<ScanListener> scanListeners = model.getScanListeners();
			synchronized(obs) {
				boolean maxChange;
				maxChange = obs.length[col][row]>obs.maxLength;
				obs.maxLengthChanged |= maxChange;
				if(maxChange)
					obs.maxLength = obs.length[col][row];
				maxChange = obs.reflCnt[col][row]>obs.maxReflCnt;
				obs.maxReflCntChanged |= maxChange;
				if(maxChange)
					obs.maxReflCnt = obs.reflCnt[col][row];
				maxChange = obs.obs1[col][row]>obs.maxObs1;
				obs.maxObs1Changed |= maxChange;
				if(maxChange)
					obs.maxObs1 = obs.obs1[col][row];
				for(ScanListener sl:scanListeners)
					sl.update(col,row);
				if(++processed==maxPos||stop)
					obs.notify();
			}			
		}
	}
}