/**
	Copyright (c) 2009-2013 by Zoltan Palmer

	This file is part of FroZenLight.

    FroZenLight is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    FroZenLight is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with FroZenLight.  If not, see http://www.gnu.org/licenses/
*/
package mvc;
import java.awt.Color;
import java.awt.EventQueue;
import java.awt.Graphics;
import java.awt.Insets;
import java.awt.Rectangle;
import java.awt.event.ComponentAdapter;
import java.awt.event.ComponentEvent;
import java.awt.event.KeyAdapter;
import java.awt.event.KeyEvent;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.awt.image.BufferedImage;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;

import javax.swing.JPanel;

import lattice.Box2D;
import lattice.Circle;
import lattice.IShape;
import lattice.Ray;
import math.Util;
import math.VecMath;
import mvc.var.VariableVisualizer;
import pattern.BaseLightPattern;

import common.ColorTheme;
import common.FloatConverter;
import common.Globals;
import common.ICallback;
import common.Log;

@SuppressWarnings("serial")
public class Visualizer<T extends Number> extends JPanel {

	public final static short LS_1 = 0;
	public final static short LS_2 = 1;
	public final static short LS_3 = 2;
	public final static short LS_4 = 3;
	public final static short[] LIGHT_SOURCES = 
		new short[]{LS_1,LS_2,LS_3,LS_4};
	private final Controller<T> ctrl;
	private final LinkedList<Short> paintedPts;
	private final LinkedList<Short> paintedShapeBounds;
	private final LightSourceSprite lightSrc;
	
	private BufferedImage bimg;
	private PaintRequestQueue queue;
	private PixelMapAbstract<T> trafo;
	private PixelMapAbstract<Float> trafoFloat;
	private ColorTheme theme;
	private boolean applPaint;
	private final ICallback callback;
	private BufferedImage circleImg;
	private Box2D<Float> canvasBox;
	private IFZLBorder border;
	private TouchButtonsOverlay<T> tbo;
	public Visualizer(Controller<T> c,
			ColorTheme ct,ICallback callback,boolean scrs) {
		this.ctrl = c;
		this.paintedPts = new LinkedList<Short>();
		this.paintedShapeBounds = new LinkedList<Short>();
		this.theme = ct;
		this.lightSrc = new LightSourceSprite();
		this.callback = callback;
		VecMath<Float> fm = ctrl.getModel().getFloatMath();
		canvasBox = new Box2D<Float>(new Float[]{0f,0f},1f,fm);
		setOpaque(true);
		setFocusable(true);
		setDoubleBuffered(false);
		border = new SimpleBorder(10,theme.getBorderColor());
		initListeners(scrs);
	}
	private Box2D<Float> getCanvas(){
		if(tbo!=null)
			return tbo.getCanvas();
		return canvasBox;
	}
	private void initListeners(boolean screensaver) {
		final Model<T> model = ctrl.getModel();
		IModelListener ml = new UpdateModelListener();
		model.addModelListener(ml);
		addComponentListener(new ResizeListener());
		if(screensaver){
			addKeyListener(new KeyAdapter() {
				@Override
				public void keyReleased(KeyEvent e) {
					System.exit(0);
				}
			});
			addMouseListener(new MouseAdapter(){
				@Override
				public void mouseReleased(MouseEvent me){
					System.exit(0);
				}
			});
			return;
		} 
		SelectRayMouseListener srml = new SelectRayMouseListener();
		addMouseListener(srml);
		addMouseMotionListener(srml);
		SelectRayKeyListener srkl = new SelectRayKeyListener();
		addKeyListener(srkl);
		RotateGridKeyListener rgkl = new RotateGridKeyListener();
		addKeyListener(rgkl);
		RotateGridListener rgl = new RotateGridListener();
		addMouseListener(rgl);
		addMouseMotionListener(rgl);
		addKeyListener(new KeyAdapter() {

			@Override
			public void keyReleased(KeyEvent e) {
				if(e.isControlDown()&&e.getKeyCode()==KeyEvent.VK_X)
					System.exit(0);
				if(e.getKeyCode()==KeyEvent.VK_ESCAPE)
					System.exit(0);
			}
		});
	}
	@Override
	public void paintComponent(Graphics g) {
		Insets insets = border.getBorderInsets(this);
		short borderX = (short) insets.left;
		short borderY = (short) insets.top;
		short w = (short)getWidth();
		short h = (short)getHeight();
		getPixelMap().setBorder(borderX,borderY);
		getFloatPixelMap().setBorder(borderX,borderY);
		getPixelMap().setSreenSize(w,h);
		getFloatPixelMap().setSreenSize(w,h);
		Float[] lgs = new Float[]{1f*(w-2*borderX-1),1f*(h-2*borderY-1)};
		getCanvas().setLengths(lgs);
		if(bimg==null||bimg.getWidth()!=w||bimg.getHeight()!=h) {
			lightSrc.valid = false;
			bimg = new BufferedImage(w,h,BufferedImage.TYPE_INT_RGB);
			Graphics ig = bimg.getGraphics();
			paintBackground(ig,w,h,borderX,borderY);
			Iterator<Short> iter = paintedShapeBounds.iterator();
			while(iter.hasNext()) {
				short x = iter.next();
				short y = iter.next();
				short width = iter.next();
				short height = iter.next();
				ig.fillOval(x,y,width,height);
			}
			paintReflections(ig,paintedPts,theme.getRayColor());
			if(pattern!=null){
				short[] defStartEnd = findDefPath(pattern);
				if(defStartEnd[0]!=-1&&defStartEnd[1]<=paintedPts.size())
					paintReflections(g,paintedPts.subList(defStartEnd[0],defStartEnd[1]),new Color(255,43,82));
			}
			ig.translate(-borderX,-borderY);
			border.drawBorder(this,ig);
		}
		if(applPaint) {
			lazyQueue();
			if(queue.size()>0) {
				PaintRequest p = queue.poll();
				doPaint(bimg.getGraphics(),p,borderX,borderY);
			}
		}
		g.drawImage(bimg,0,0,this);
	}
	private void paintBackground(Graphics g,int w,int h,int bx,int by){
		if(tbo!=null){
			tbo.paintBackround(g,w,h,bx,by,theme.getBgColor());
			tbo.paintButtons(g,w,h,bx,by);
		} else {
			g.translate(bx,by);
			g.setColor(theme.getBgColor());
			g.fillRect(0,0,w-2*bx,h-2*by);
		}
	}

	private void lazyQueue(){
		synchronized(callback){
			if(queue==null) {
				queue = new PaintRequestQueue();
				callback.call();
			}
		}
	}
	private void doPaint(Graphics g,PaintRequest req,short bx,short by) {
		if(req.background) {
			border.drawBorder(this,g);
			paintBackground(g,getWidth(),getHeight(),bx,by);
			paintedPts.clear();
			paintedShapeBounds.clear();
			lightSrc.valid = false;
		} else {
			g.translate(bx,by);
			if(req.newray) {
				lightSrc.delete();
				paintReflections(g,paintedPts,theme.getBgColor());
				paintedPts.clear();
			}
			if(req.circles||req.newray) {
				lightSrc.delete();
				g.setColor(theme.getBgColor());
				Iterator<Short> iter = paintedShapeBounds.iterator();
				while(iter.hasNext()) {
					short x = iter.next();
					short y = iter.next();
					short w = iter.next();
					short h = iter.next();
					g.fillRect(x,y,w,h);
				}
				paintedShapeBounds.clear();
			}
		}
		if(paintedShapeBounds.size()==0)
			paintShapes(g,req.iter);
		if(req.data!=null&&req.data.size()>0) {
			Iterator<T> iter = req.data.iterator();
			while(iter.hasNext()) {
				float x = iter.next().floatValue();
				float y = iter.next().floatValue();
				short x0 = getFloatPixelMap().xWorld2Pix(x);
				short y0 = getFloatPixelMap().yWorld2Pix(y);
				if(paintedPts.size()==0){
					// new code
					if(tbo!=null){
						Float[] pt2 = new Float[]{1f*x0,1f*y0};
						if(!getCanvas().containsLocal(pt2)){
							Float[] dir = new Float[]{-1f,0f};
							Ray<Float> ray = new Ray<Float>(pt2.clone(),dir,null);
							if(getCanvas().intersect(ray,pt2))
								x0 = (short)(pt2[0]-(lightSrc.width()-lightSrc.xOffset()));
							else{
								dir[0] = 1f;
								getCanvas().intersect(ray,pt2);
								x0 = (short)(pt2[0]+(lightSrc.width()-lightSrc.xOffset()));
							}
						}						
					}else{// new code
						int x1 = lightSrc.width()-lightSrc.xOffset();
						x0 = (short)Math.min(getWidth()-2*bx-x1,x0);
					}
				} else {
					Float[] pt2 = new Float[]{1f*x0,1f*y0};
					if(!getCanvas().containsLocal(pt2)){
						short xp = paintedPts.get(paintedPts.size()-2);
						short yp = paintedPts.get(paintedPts.size()-1);
						Float[] pt1 = new Float[]{1f*xp,1f*yp};
						Float[] dir = new Float[]{pt2[0]-pt1[0],pt2[1]-pt1[1]};
						ctrl.getModel().getFloatMath().normalize(dir);
						Ray<Float> ray = new Ray<Float>(pt1,dir,null);
						getCanvas().intersect(ray,pt2);
						x0 = pt2[0].shortValue();
						y0 = pt2[1].shortValue();
					}
				}
				paintedPts.add(x0);
				paintedPts.add(y0);
			}
			if(req.newray)	
				lightSrc.put(paintedPts.get(0),paintedPts.get(1));
			paintReflections(g,paintedPts,theme.getRayColor());
			if(pattern!=null){
				short[] defStartEnd = findDefPath(pattern);
				if(defStartEnd[0]!=-1&&defStartEnd[1]<=paintedPts.size())
					paintReflections(g,paintedPts.subList(defStartEnd[0],defStartEnd[1]),new Color(255,43,82));
			}
			lightSrc.paint(g);
		}
		g.translate(-bx,-by);
	}
	public PixelMapAbstract<Float> getFloatPixelMap() {
		if(trafoFloat==null) {
			Model<T> model = ctrl.getModel();
			Box2D<T> worldBox = model.getWorldBox();
			VecMath<Float> fmath = model.getFloatMath();
			FloatConverter<T> tfltCvtr = new FloatConverter<T>();
			Box2D<Float> box = worldBox.convert(tfltCvtr,fmath);
			Insets insets = border.getBorderInsets(this);
			short bx = (short) insets.left;
			short by = (short) insets.top;
			trafoFloat = new PixelMapFloat(getWidth(),getHeight(),bx,by,box);
		}
		return trafoFloat;
	}
	PixelMapAbstract<T> getPixelMap() {
		if(trafo==null) {
			Model<T> model = ctrl.getModel();
			Box2D<T> box = model.getWorldBox().copy();
			VecMath<T> mt = model.getMath();
			Insets insets = border.getBorderInsets(this);
			short bx = (short) insets.left;
			short by = (short) insets.top;
			trafo = new PixelMap<T>(getWidth(),getHeight(),bx,by,box,mt);
		}
		return trafo;
	}
	private void paintShapes(Graphics g,Iterator<IShape<Float>> it) {
		short[] pixPt = new short[2];
		Float[] wPt = new Float[2];
		PixelMapAbstract<Float> pixMap = getFloatPixelMap();
		while(it.hasNext()) {
			IShape<Float> shape = it.next();
			wPt[0] = shape.getCenter()[0];
			wPt[1] = shape.getCenter()[1];
			if(shape instanceof Circle) {
				float r = ((Circle<Float>)shape).getRadius();
				wPt[0] -= r;
				wPt[1] += r;
				pixMap.world2Pix(wPt,pixPt);
				short r2Pix = pixMap.xLengthWorld2Pix(2*r);
				fillCircle(g,pixPt[0],pixPt[1],r2Pix);
				paintedShapeBounds.add(pixPt[0]);
				paintedShapeBounds.add(pixPt[1]);
				paintedShapeBounds.add(r2Pix);
				paintedShapeBounds.add(r2Pix);
			}
		}
	}
	void requestDefPath(BaseLightPattern<T> b){	pattern = b; }
	BaseLightPattern<T> pattern;
	short[] findDefPath(BaseLightPattern<T> blp){
		List<short[]> bcs = blp.getCellSequence();
		if(blp.isPropagating())
			return new short[]{2,(short) (2+2*bcs.size())};
		ArrayList<Short> cells = new ArrayList<Short>();
		short s = ctrl.getModel().getGridSize();
		for(int i=2;i<paintedPts.size()-2;i+=2){
			for(int j=0;j<paintedShapeBounds.size();j+=4){
				Short x = paintedPts.get(i);
				Short y = paintedPts.get(i+1);
				float r = paintedShapeBounds.get(j+2)/2;
				float a = paintedShapeBounds.get(j)+r;
				float b = paintedShapeBounds.get(j+1)+r;
				if((x-a)*(x-a)+(y-b)*(y-b)<=(r+3)*(r+3)){
					cells.add((short)(j/(4*s)));
					cells.add((short)((j/4)%s));
					break;
				}
			}
			if(cells.size()!=i)
				return new short[]{-1,-1};
		}
		int i;
		boolean[] ok = new boolean[]{false,false,false,false};
		int idx = 1;
		for(i=0;i<cells.size()-2*bcs.size()&&!(ok[0]||ok[1]||ok[2]||ok[3]);i+=2){
			//for(idx = 0;idx<2&&!(ok1||ok2||ok3||ok4);idx++){
				ok = new boolean[]{true,true,true,true};
				for(int j=bcs.size()-1;j>=0&&(ok[0]||ok[1]||ok[2]||ok[3]);j--){
					Short cx = cells.get(i+2*j);
					Short cy = cells.get(i+1+2*j);
					if(ok[0])
						ok[0] = (cx==bcs.get(j)[idx])&&(cy==bcs.get(j)[1-idx]);
					if(ok[1])
						ok[1] = (cx==s-1-bcs.get(j)[idx])&&(cy==bcs.get(j)[1-idx]);
					if(ok[2])
						ok[2] = (cx==bcs.get(j)[idx])&&(cy==s-1-bcs.get(j)[1-idx]);
					if(ok[3])
						ok[3] = (cx==s-1-bcs.get(j)[idx])&&(cy==s-1-bcs.get(j)[1-idx]);
				}
			//}
		}
		if(ok[0]||ok[1]||ok[2]||ok[3])
			return new short[]{(short)(i),(short)(i+bcs.size()*2)};
		return new short[]{-1,-1};
	}
	private void fillCircle(Graphics g,short ptx,short pty,short s){
		if(circleImg==null||circleImg.getHeight()!=s){
			circleImg = new BufferedImage(s,s,BufferedImage.TYPE_INT_ARGB);
			refreshCircleImg();
		}
		g.drawImage(circleImg,ptx,pty,null);
	}
	private void refreshCircleImg(){
		Graphics ig = circleImg.getGraphics();
		//ig.setColor(theme.getBgColor());
		ig.setColor(new Color(0,0,0,0));
		int s = circleImg.getHeight();
		ig.fillRect(0,0,s,s);
		ig.setColor(theme.getShapeColor());
		ig.fillOval(0,0,s,s);
	}
	private static void paintReflections(Graphics g,List<Short> pts,Color color) {
		Iterator<Short> iter = pts.iterator();
		if(iter.hasNext()) {
			g.setColor(color);
			short startx = iter.next();
			short starty = iter.next();
			while(iter.hasNext()) {
				short endx = iter.next();
				short endy = iter.next();
				g.drawLine(startx,starty,endx,endy);
				startx = endx;
				starty = endy;
			}
		}
	}
	public void draw(LinkedList<T> coords,Iterator<IShape<Float>> it,
			boolean c,boolean r) {
		//if(queue==null)
		//	return;
		//lazyQueue();
		queue.push(new PaintRequest(coords,false,c,r,it));
		EventQueue.invokeLater(new Runnable() {

			@Override
			public void run() {
				paintImmediately(getVisibleRect());
			}
		});
	}
	@Override
	public void paintImmediately(int x,int y,int w,int h) {
		applPaint = true;
		super.paintImmediately(x,y,w,h);
		applPaint = false;
	}
	@Override
	public void paintImmediately(Rectangle r) {
		applPaint = true;
		super.paintImmediately(r);
		applPaint = false;
	}
	@Override
	public void repaint() {
		if(ctrl!=null) {
			if(queue!=null)
				queue.push(createPaintAllRequest());
			EventQueue.invokeLater(new Runnable() {

				@Override
				public void run() {
					paintImmediately(getVisibleRect());
				}
			});
		}
	}
	public PaintRequest createPaintAllRequest() {
		LinkedList<T> data = new LinkedList<T>();
		Model<T> model = ctrl.getModel();
		model.getRayDataComplete(data);
		return new PaintRequest(data,true,true,true,model.getShapeIteratorFloat());
	}
	public void setTheme(ColorTheme theme) {
		this.theme = theme;
		if(ctrl.getModel().getShape() instanceof Circle)
			if(circleImg!=null)
				refreshCircleImg();
		lightSrc.refresh();
		border.setColor(theme.getBorderColor());
		if(tbo!=null)
			tbo.setColor(getTheme().getBorderColor());
	}
	public ColorTheme getTheme() {
		return theme;
	}
	public int getQueueSize() {
	    return queue.size();
    }
	public void setLightSourceType(short lsType){
		lightSrc.setType(lsType);
		repaint();		
	}
	public void setBorder(IFZLBorder b) {
		border = b;
		new ResizeListener().componentResized(new ComponentEvent(this,0));
		repaint();
	}
	private final class LightSourceSprite {
		
		private final short[] pos;
		private final int[][] pixBuf;
		private List<Byte> xoffsets;
		private List<byte[][]> defs; 
		private BufferedImage img;
		private boolean valid;
		private short shape; 
		LightSourceSprite() {
			pos = new short[2];
			pixBuf = new int[10][9];
			valid = false;
			shape = LIGHT_SOURCES[0];
			xoffsets = new ArrayList<Byte>();
			defs = new ArrayList<byte[][]>();
			defs.add(new byte[][]{
						new byte[]{0,0,2,2,2,2,2,2,2,0},
						new byte[]{0,0,2,1,1,1,1,1,2,2},
						new byte[]{0,0,2,2,2,2,2,2,1,2},
						new byte[]{0,0,0,2,1,1,1,2,1,2},
						new byte[]{0,0,0,0,1,1,1,2,1,0},
						new byte[]{0,0,0,2,1,1,1,2,1,2},
						new byte[]{0,0,2,2,2,2,2,2,1,2},
						new byte[]{0,0,2,1,1,1,1,1,2,2},
						new byte[]{0,0,2,2,2,2,2,2,2,0}});
			xoffsets.add((byte)4);
			defs.add(new byte[][]{
						new byte[]{0,0,0,0,0,0,0,0,0,0},
						new byte[]{0,0,2,2,2,2,2,2,2,0},
						new byte[]{2,2,2,1,2,1,2,1,2,2},
						new byte[]{2,1,1,2,1,2,1,2,1,2},
						new byte[]{0,1,1,2,1,2,1,2,1,0},
						new byte[]{2,1,1,2,1,2,1,2,1,2},
						new byte[]{2,2,2,1,2,1,2,1,2,2},
						new byte[]{0,0,2,2,2,2,2,2,2,0},
						new byte[]{0,0,0,0,0,0,0,0,0,0}});
			xoffsets.add((byte)1);
			defs.add(new byte[][]{
						new byte[]{0,0,0,0,0,0,0,0,0,2},
						new byte[]{2,2,2,2,2,2,2,2,2,2},
						new byte[]{2,1,1,1,1,1,1,1,1,2},
						new byte[]{2,1,1,2,2,2,2,2,1,2},
						new byte[]{0,1,1,0,0,0,0,0,1,0},
						new byte[]{2,1,1,2,2,2,2,2,1,2},
						new byte[]{2,1,1,1,1,1,1,1,1,2},
						new byte[]{2,2,2,2,2,2,2,2,2,2},
						new byte[]{0,0,0,0,0,0,0,0,0,2}});
			xoffsets.add((byte)1);
			setType((short)0);
		}
		void setType(short t){
			shape = t;
			int lastIdx = LIGHT_SOURCES.length-1;
			if(shape!=LIGHT_SOURCES[lastIdx]){
				img = new BufferedImage(10,9,BufferedImage.TYPE_INT_ARGB);
				Graphics ig = img.getGraphics();
				Color trans = new Color(0,0,0,0);
				byte idx = Globals.getIndex(shape,LIGHT_SOURCES);
				byte[][] def = defs.get(idx);
				for(int i = 0;i<def.length;i++)
					for(int k = 0;k<def[0].length;k++)
						if(def[i][k]==0){
							ig.setColor(trans);
							ig.drawRect(k,i,0,0);
						} else if(def[i][k]==1) {
							ig.setColor(theme.getRayColor().darker());
							ig.drawRect(k,i,0,0);
						}
				drawProtectedRegion();
			}
		}
		void delete() {
			if(valid) {
				Insets insets = border.getBorderInsets(Visualizer.this);
				short bx = (short) insets.left;
				short by = (short) insets.top;
				short px = (short)(pos[0]+bx);
				short py = (short)(pos[1]+by);
				for(int i = 0;i<pixBuf.length;i++)
					for(int k = 0;k<pixBuf[0].length;k++)
						bimg.setRGB(px+i-xOffset(),py-4+k,pixBuf[i][k]);
				valid = false;
			}
		}
		void put(short x,short y) {
			Insets insets = border.getBorderInsets(Visualizer.this);
			short bx = (short) insets.left;
			short by = (short) insets.top;
			pos[0] = (short)Math.min(getWidth()-2*bx-10+xOffset(),x);
			pos[1] = (short)Math.min(getHeight()-2*by,y);
			short px = (short)(pos[0]+bx);
			short py = (short)(pos[1]+by);
			for(int i = 0;i<pixBuf.length;i++)
				for(int k = 0;k<pixBuf[0].length;k++)
					pixBuf[i][k] = bimg.getRGB(px+i-xOffset(),py-4+k);
			drawProtectedRegion();
			valid = true;
		}
		void refresh(){ setType(shape); }
		int xOffset(){
			byte idx = Globals.getIndex(shape,LIGHT_SOURCES);
			return idx<xoffsets.size()?xoffsets.get(idx):0;
		}
		int width(){ return pixBuf.length; }
		private void drawProtectedRegion(){
			byte idx = Globals.getIndex(shape,LIGHT_SOURCES);
			if(idx!=LIGHT_SOURCES.length-1){
				byte[][] def = defs.get(idx);
				Graphics ig = img.getGraphics();
				for(int i = 0;i<def.length;i++)
					for(int k = 0;k<def[0].length;k++){
						if(def[i][k]==2){
							ig.setColor(new Color(pixBuf[k][i]));
							ig.drawRect(k,i,0,0);
						}
					}
			}
		}
		boolean hit(short[] xy) {
			short dx = (short)(xy[0]-pos[0]);
			short dy = (short)(xy[1]-pos[1]);
			return dx*dx+dy*dy<400;
		}
		void paint(Graphics g) {
			int lastIdx = LIGHT_SOURCES.length-1;
			if(shape!=LIGHT_SOURCES[lastIdx])
				g.drawImage(img,pos[0]-xOffset(),pos[1]-4,null);
		}
	}
	private final class UpdateModelListener extends IModelListener {

		private final LinkedList<T> data;
		boolean newCircleDraw = true;
		boolean newRayDraw = false;
		UpdateModelListener() {
			this.data = new LinkedList<T>();
		}
		@Override
		public void modelChanged(byte event) {
if(event!=Model.MODEL_WORLD_RESIZE){
				//defStartEnd[0] = -1;
				//defStartEnd[1] = -1;
pattern = null;
}
			switch(event) {
				case Model.MODEL_WORLD_RESIZE:
					trafo = null;
					trafoFloat = null;
					// Repaint is already requested by grid bag layout
					// repaint();
					break;
				case Model.MODEL_NEW:
					trafo = null;
					trafoFloat = null;
					repaint();
					break;
				case Model.RAY_NEW:
					newRayDraw = true;
					break;
				case Model.GRID_RESIZE:
					newCircleDraw = true;
					newRayDraw = true;
					break;
				case Model.GRID_ROTATE:
					newCircleDraw = true;
					newRayDraw = true;
					break;
				case Model.SHAPE_RESIZE:
					newCircleDraw = true;
					newRayDraw = true;
					if(tbo!=null)
						repaint();
					break;
				case Model.RAY_UPDATE:
					if(queue==null){
						newCircleDraw = true;
						newRayDraw = true;						
					} else {
						data.clear();
						Model<T> model = ctrl.getModel();
						model.getRayDataUpdate(data);
						Iterator<IShape<Float>> shapeIter = null;
						if(newCircleDraw||newRayDraw)
							shapeIter = model.getShapeIteratorFloat();
						draw(data,shapeIter,newCircleDraw,newRayDraw);
						newCircleDraw = false;
						newRayDraw = false;
					}
					break;
			}
		}
	}
	private final class RotateGridListener extends MouseAdapter {

		private Float[] pt1;
		private boolean ignore;
		@Override
		public void mouseDragged(MouseEvent e) {
			if(!ignore&&!ctrl.isRotAnimRun()&&!ctrl.isPatAnimRun()) {
				short[] pt = getPixelMap().toCanvas((short)e.getX(),(short)e.getY());
				if(pt1==null&&lightSrc.hit(pt))
					ignore = true;
				else {
					PixelMapAbstract<Float> pixMap = getFloatPixelMap();
					short[] cvpt = pixMap.toCanvas((short)e.getX(),(short)e.getY());
					Float[] pt2 = new Float[2];
					pixMap.pix2World(cvpt,pt2);
					Model<T> model = ctrl.getModel();
					Util<T> util = model.getMath().util;
					if(model.isInGrid(util.cast(pt2))) {
						if(pt1!=null)
							rotateGrid(pt1,pt2,Model.DYNAMIC_CALC);
						pt1 = pt2;
					} else if(pt1==null)
						ignore = true;
				}
			}
		}
		@Override
		public void mouseReleased(MouseEvent e) {
			if(pt1!=null) {
				PixelMapAbstract<Float> pixMap = getFloatPixelMap();
				short[] cvpt = pixMap.toCanvas((short)e.getX(),(short)e.getY());
				Float[] pt2 = new Float[2];
				pixMap.pix2World(cvpt,pt2);
				rotateGrid(pt1,pt2,Model.STATIC_CALC);
			}
			pt1 = null;
			ignore = false;
		}
		private void rotateGrid(Float[] p1,Float[] p2,byte mode) {
			ctrl.stopCalculation();
			float cosphi = p1[0]*p2[0]+p1[1]*p2[1];
			cosphi /= Math.sqrt(p1[0]*p1[0]+p1[1]*p1[1]);
			cosphi /= Math.sqrt(p2[0]*p2[0]+p2[1]*p2[1]);
			cosphi = Math.max(Math.min(cosphi,1),-1);
			float s = Math.signum(p1[0]*p2[1]-p1[1]*p2[0]);
			float phi = (float)(s*Math.acos(cosphi));
			float pifactor = (float)(phi/Math.PI);
			VecMath<T> math = ctrl.getModel().getMath();
			T[] sincosphi = math.discrSinCos(pifactor);
			T discrPhi = math.discrAngle(pifactor);
			ctrl.waitForCalculatorRelease();
			ctrl.getModel().setCalcMode(mode);
			ctrl.getModel().rotateGrid(discrPhi,sincosphi);
			ctrl.startCalculation();
		}
	}
	private final class RotateGridKeyListener extends KeyAdapter {

		@Override
		public void keyReleased(KeyEvent e) {
			if(!ctrl.isRotAnimRun()&&!ctrl.isPatAnimRun()){
				int kc = e.getKeyCode();
				if(kc==KeyEvent.VK_LEFT||kc==KeyEvent.VK_RIGHT) {
					Model<T> model = ctrl.getModel();
					ctrl.stopCalculation();
					VecMath<T> math = model.getMath();
					float dphi = e.getKeyCode()==KeyEvent.VK_RIGHT?-0.001f:0.001f;
					T[] sincosphi = math.discrSinCos(dphi);
					T phi = math.discrAngle(dphi);
					ctrl.waitForCalculatorRelease();
					// TODO Check for key buffer
					model.setCalcMode(Model.STATIC_CALC);
					model.rotateGrid(phi,sincosphi);
					ctrl.startCalculation();
				}
			}
		}
	}
	private final class TouchButtonsMouseListener extends MouseAdapter {
		boolean releaseRepaint = false;
		public void mousePressed(MouseEvent e) {
			Insets insets = border.getBorderInsets(Visualizer.this);
			short bx = (short) insets.left;
			short by = (short) insets.top;
			if(tbo.handleButtonPress(getWidth(),getHeight(),bx,by,e.getX()-bx,e.getY()-by)){
				//paintImmediately(0,0,getWidth(),getHeight());
				repaint();
				releaseRepaint = true;
			}
		}
		public void mouseReleased(MouseEvent e){
			if(releaseRepaint){
				repaint();
				//paintImmediately(0,0,getWidth(),getHeight());
				releaseRepaint = false;
			}
		}
	}
	private final class SelectRayMouseListener extends MouseAdapter {

		private short[] pt;
		public SelectRayMouseListener() {}
		@Override
		public void mouseDragged(MouseEvent e) {
			if(pt!=null){
				short[] pt2 = getPixelMap().toCanvas((short)e.getX(),(short)e.getY());
				setRay(pt2,Model.DYNAMIC_CALC);
			}
		}
		@Override
		public void mousePressed(MouseEvent e) {
			if(!ctrl.isPatAnimRun()){
				pt = getPixelMap().toCanvas((short)e.getX(),(short)e.getY());
				if(!lightSrc.hit(pt))
					pt = null;
			}
		}
		@Override
		public void mouseReleased(MouseEvent e) {
			if(pt!=null){
				short[] pt2 = getPixelMap().toCanvas((short)e.getX(),(short)e.getY());
				if(pt[0]-pt2[0]!=0||pt[1]-pt2[1]!=0)
					setRay(pt2,Model.STATIC_CALC);
				pt = null;
			}
		}
		private void setRay(short[] pt,byte mode) {
			if(!ctrl.isRotAnimRun())
				ctrl.stopCalculation();
			Model<T> model = ctrl.getModel();
			getPixelMap().clip2Canvas(pt,(short) 1);
			VecMath<T> mt = model.getMath();
			T[] wp = mt.util.array(2);
			getPixelMap().pix2World(pt[0],pt[1],wp);
			wp[0] = mt.util.round(wp[0],VariableVisualizer.DIST_DIGITS);
			wp[0] = mt.util.expand(wp[0],mt.PRECISION);
			T[] dir = model.LIGHT_SOURCE_DIR;
			Ray<T> ray = new Ray<T>(wp,dir,model.getMath().util);
			if(!ctrl.isRotAnimRun()){
				ctrl.waitForCalculatorRelease();
				model.setRay(ray,mode);
				ctrl.startCalculation();
			} else 
				ctrl.setRayLater(ray);
		}
	}
	private final class SelectRayKeyListener extends KeyAdapter {

		public SelectRayKeyListener() {}
		@Override
		public void keyReleased(KeyEvent e) {
			if(ctrl.isPatAnimRun())
				return;
			if(e.getKeyCode()==KeyEvent.VK_UP||e.getKeyCode()==KeyEvent.VK_DOWN) {
				boolean upDown = e.getKeyCode()==KeyEvent.VK_UP;
				Model<T> model = ctrl.getModel();
				if(!ctrl.isRotAnimRun())
					ctrl.stopCalculation();
				T[] start = model.getRay().getStart().clone();
				T[] dir = model.LIGHT_SOURCE_DIR;
				Ray<T> ray = new Ray<T>(start,dir,model.getMath().util);
				PixelMapAbstract<T> pixMap = getPixelMap();
				T dy = pixMap.yLengthPix2World((short)1);
				if(upDown)
					start[1] = model.getMath().add(start[1],dy);
				else
					start[1] = model.getMath().sub(start[1],dy);
				// TODO Info about existence of key buffer is needed
				if(!ctrl.isRotAnimRun()){
					ctrl.waitForCalculatorRelease();
					model.setRay(ray,Model.STATIC_CALC);
					ctrl.startCalculation();
				} else 
					ctrl.setRayLater(ray);
			}
		}
	}
	private final class ResizeListener extends ComponentAdapter {
		@Override
		public void componentResized(ComponentEvent ce) {
			boolean stop = ctrl.stopCalculation();
			Model<T> model = ctrl.getModel();
			VecMath<T> mt = model.getMath();
			Insets ins = border.getBorderInsets(Visualizer.this);
			short borderX = (short) ins.left;
			short borderY = (short) ins.top;
			ctrl.waitForCalculatorRelease();
			short w = (short)getWidth();
			short h = (short)getHeight();
//			if(w>2*borderX&&h>2*borderY){
				getPixelMap().setBorder(borderX,borderY);
				getFloatPixelMap().setBorder(borderX,borderY);
				getPixelMap().setSreenSize(w,h);
				getFloatPixelMap().setSreenSize(w,h);
//			}
			Box2D<T> world = model.getWorldBox();
			Box2D<T> newWorld = ctrl.createNewWorld(w-2*borderX-1,h-2*borderY-1,mt);
			world.setCorner(newWorld.getCorner());
			if(model.getRay()!=null){
				if(model.updateWorldBox(newWorld))
					ctrl.startCalculation();
				else if(stop){
					String descr = model.getDescription();
					model.setRay(model.getRay(),descr,Model.STATIC_CALC);
					ctrl.startCalculation();
				}
			}
		}
	}
	private final class PaintRequestQueue {

		private final ArrayList<PaintRequest> requests = new ArrayList<PaintRequest>();
		public synchronized PaintRequest poll() {
			if(requests.size()>0) 
				return requests.remove(requests.size()-1);
			return null;
		}
		public synchronized void push(PaintRequest request) {
			requests.add(0,request);
		}
		public int size() {
			return requests.size();
		}
	}
	private final class PaintRequest {

		private LinkedList<T> data;
		boolean background;
		boolean circles;
		boolean newray;
		Iterator<IShape<Float>> iter;
		public PaintRequest(LinkedList<T> d,boolean b,boolean c,boolean r,
				Iterator<IShape<Float>> it) {
			if(d!=null) {
				this.data = new LinkedList<T>(d);
				if(Globals.LOG_VERBOSE)
					if(d.size()==0)
						Log.getIstc().logln("WARNING: Empty data paint request!");
			}
			this.background = b;
			this.circles = c;
			this.newray = r;
			this.iter = it;
		}
		@Override
		public String toString() {
			return "bg:"+background+" cr:"+circles+" nr:"+newray+" sz:"+data.size();
		}
	}
	public void setTouchOverlay(TouchButtonsOverlay touchButtonsOverlay) {
		this.tbo = touchButtonsOverlay;
		tbo.setColor(getTheme().getBorderColor());
		TouchButtonsMouseListener tbml = new TouchButtonsMouseListener();
		addMouseListener(tbml);
		setBorder(new SimpleBorder(50,theme.getBorderColor()));
	}
}
