/**
	Copyright (c) 2009-2013 by Zoltan Palmer

	This file is part of FroZenLight.

    FroZenLight is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    FroZenLight is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with FroZenLight.  If not, see http://www.gnu.org/licenses/
*/
package mvc.mgr;
import java.awt.Component;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.event.ItemEvent;
import java.awt.event.ItemListener;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.util.ArrayList;
import java.util.List;

import javax.swing.BorderFactory;
import javax.swing.DefaultCellEditor;
import javax.swing.JComboBox;
import javax.swing.JMenuItem;
import javax.swing.JPanel;
import javax.swing.JPopupMenu;
import javax.swing.JScrollPane;
import javax.swing.JTable;
import javax.swing.JTextField;
import javax.swing.ListSelectionModel;
import javax.swing.ScrollPaneConstants;
import javax.swing.SwingUtilities;
import javax.swing.border.TitledBorder;
import javax.swing.event.CellEditorListener;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ListSelectionEvent;
import javax.swing.event.ListSelectionListener;
import javax.swing.table.DefaultTableCellRenderer;
import javax.swing.table.TableCellEditor;
import javax.swing.table.TableColumn;

import math.VecMath;
import pattern.BaseLightPattern;

import common.Globals;

@SuppressWarnings("serial")
public class ArtLightPatternEditor<T extends Number>
        extends BaseLightPatternEditor<T,BaseLightPattern<T>> {

	private final static String[] BND_VAL_TYPES;
	private final static String[] BND_VAL_POINT;
	private final static String[] BND_VAL_GRAD;
	static {
		BND_VAL_TYPES = new String[]{"POINT","GRADIENT LEFT","GRADIENT RIGHT"};
		BND_VAL_POINT = new String[]{"FIRST","SECOND"};
		BND_VAL_GRAD = new String[]{"-1","0","1"};
	}
	private final static String INFINITY = "inf";
	private JComboBox startBTComboBox;
	private JComboBox startBVComboBox;
	private JComboBox endBTComboBox;
	private JComboBox endBVComboBox;
	private SeqenceTableModel seqTableModel;
	private JTable seqTable;
	private JScrollPane seqTableScroller;
	public ArtLightPatternEditor(VecMath<T> mt) {
		super(true,2,mt);
	}
	@Override
	void setLightPattern(BaseLightPattern<T> lp) {
		setLightPatternName(lp.getName());
		setGridSize(lp.getGridSize());
		setMaxRadius(lp.getMaxShapeSize());
		switch(lp.getStartBndValType()) {
			case BaseLightPattern.POINT:
				startBTComboBox.setSelectedItem(BND_VAL_TYPES[0]);
				break;
			case BaseLightPattern.GRAD_LEFT:
				startBTComboBox.setSelectedItem(BND_VAL_TYPES[1]);
				break;
			case BaseLightPattern.GRAD_RIGHT:
				startBTComboBox.setSelectedItem(BND_VAL_TYPES[2]);
		}
		switch(lp.getEndBndValType()) {
			case BaseLightPattern.POINT:
				endBTComboBox.setSelectedItem(BND_VAL_TYPES[0]);
				break;
			case BaseLightPattern.GRAD_LEFT:
				endBTComboBox.setSelectedItem(BND_VAL_TYPES[1]);
				break;
			case BaseLightPattern.GRAD_RIGHT:
				endBTComboBox.setSelectedItem(BND_VAL_TYPES[2]);
		}
		seqTableModel.clear();
		List<short[]> cells = lp.getCellSequence();
		if(cells!=null) {
			while(seqTableModel.getRowCount()<cells.size()+10)
				seqTableModel.insertRow(0);
			while(seqTableModel.getRowCount()>cells.size()+10)
				seqTableModel.removeRows(new int[]{0});
			for(int i = 0;i<cells.size();i++) {
				short[] cell = cells.get(i);
				String cellStr = " ("+cell[0]+","+cell[1]+")";
				seqTableModel.setValueAt(cellStr,i,1);
			}
			int i1 = lp.getStartBndValType()==BaseLightPattern.POINT?1:0;
			int i2 = lp.getEndBndValType()==BaseLightPattern.POINT?1:0;
			for(int i = i1;i<cells.size()-i2;i++) {
				float gr = lp.getInitialGradients().get(i-i1);
				seqTableModel.setValueAt(" "+gr,i,2);
			}
			T bv = lp.getStartBndVal();
			if(lp.getStartBndValType()==BaseLightPattern.POINT) {
				seqTableModel.setValueAt(" "+getBndValStr(bv),0,2);
				startBVComboBox.setSelectedItem(getBndValSelectStr(bv));
			} else
				startBVComboBox.setSelectedItem(bv.byteValue()+"");
			bv = lp.getEndBndVal();
			if(lp.getEndBndValType()==BaseLightPattern.POINT) {
				int row = seqTableModel.getLastRowPos();
				seqTableModel.setValueAt(" "+getBndValStr(bv),row,2);
				endBVComboBox.setSelectedItem(getBndValSelectStr(bv));
			} else
				endBVComboBox.setSelectedItem(bv.byteValue()+"");
		}
	}
	@Override
	BaseLightPattern<T> getLightPattern() throws Exception {
		stopEditing(seqTable);
		short maxPrec = Globals.PRECISIONS[Globals.PRECISIONS.length-1];
		BaseLightPattern<T> lp = new BaseLightPattern<T>(math);
		lp.setName(getLightPatternName());
		if(getGridSize()==0)
			throw new Exception("The size of the grid must be at least 2x2.");
		lp.setGridSize(getGridSize());
		if(math.cmp(getMaxRadius(),math.ZERO)==0)
			throw new Exception("The maximum radius must be > 0 %.");
		lp.setMaxShapeSize(math.util.copy(getMaxRadius(),maxPrec));
		String sbt = (String)startBTComboBox.getSelectedItem();
		if(sbt.compareTo(BND_VAL_TYPES[0])==0)
			lp.setStartBndValType(BaseLightPattern.POINT);
		else if(sbt.compareTo(BND_VAL_TYPES[1])==0)
			lp.setStartBndValType(BaseLightPattern.GRAD_LEFT);
		else
			lp.setStartBndValType(BaseLightPattern.GRAD_RIGHT);
		String ebt = (String)endBTComboBox.getSelectedItem();
		if(ebt.compareTo(BND_VAL_TYPES[0])==0)
			lp.setEndBndValType(BaseLightPattern.POINT);
		else if(ebt.compareTo(BND_VAL_TYPES[1])==0)
			lp.setEndBndValType(BaseLightPattern.GRAD_LEFT);
		else
			lp.setEndBndValType(BaseLightPattern.GRAD_RIGHT);
		if(seqTableModel.getLastRowPos()==-1)
			throw new Exception("The sequence of grid cells is missing.");
		if(lp.getStartBndValType()==BaseLightPattern.POINT) {
			String select = (String)startBVComboBox.getSelectedItem();
			String value = ((String)seqTableModel.getValueAt(0,2)).trim();
			T bv = getBoundValue(value,select);
			lp.setStartBndVal(bv);
		} else {
			byte value = Byte.parseByte((String)startBVComboBox.getSelectedItem());
			lp.setStartBndVal(math.util.copy(math.util.cast(value),maxPrec));
		}
		if(lp.getEndBndValType()==BaseLightPattern.POINT) {
			String select = (String)endBVComboBox.getSelectedItem();
			int lastRow = seqTableModel.getLastRowPos();
			String value = ((String)seqTableModel.getValueAt(lastRow,2)).trim();
			T bv = getBoundValue(value,select);
			lp.setEndBndVal(bv);
		} else {
			byte value = Byte.parseByte((String)endBVComboBox.getSelectedItem());
			lp.setEndBndVal(math.util.copy(math.util.cast(value),maxPrec));
		}
		ArrayList<short[]> cells = new ArrayList<short[]>();
		for(int row = 0;row<=seqTableModel.getLastRowPos();row++) {
			String cell = (String)seqTableModel.getValueAt(row,1);
			int i1 = cell.indexOf("(");
			int i2 = cell.indexOf(",");
			int i3 = cell.indexOf(")");
			if(i1!=-1&&i2!=-1&&i3!=-1) {
				short val1 = Short.parseShort(cell.substring(i1+1,i2));
				short val2 = Short.parseShort(cell.substring(i2+1,i3));
				cells.add(new short[]{val1,val2});
			} else {
				String txt = "Error in line "+row+":\n"+"The format of the cells is (x,y).";
				throw new Exception(txt);
			}
		}
		lp.setCellSequence(cells);
		lp.calcStartIndex();
		if(lp.getStartIndex()==-1)
			throw new Exception("All cells lie outside of the grid.");
		else if(lp.getStartIndex()!=0)
			lp.setPropagating(true);
		int i1 = lp.getStartBndValType()==BaseLightPattern.POINT?1:0;
		int i2 = lp.getEndBndValType()==BaseLightPattern.POINT?1:0;
		ArrayList<Float> gradients = new ArrayList<Float>();
		for(int row = i1;row<=seqTableModel.getLastRowPos()-i2;row++) {
			String value = ((String)seqTableModel.getValueAt(row,2)).trim();
			try {
				if(value.compareTo(INFINITY)==0)
					gradients.add(1000f);
				else
					gradients.add(Float.parseFloat(value));
			} catch(Exception e) {
				String txt = "Error in line "+row+":\n"
				        +"The gradient is neither a float number nor an "+INFINITY+".";
				throw new Exception(txt);
			}
		}
		lp.setInitialGradients(gradients);
		return lp;
	}
	private String getBndValSelectStr(T bv) {
		if(bv.byteValue()<4)
			return BND_VAL_POINT[0];
		return BND_VAL_POINT[1];
	}
	private String getBndValStr(T bv) {
		byte pos = bv.byteValue();
		if(pos>=4)
			pos -= 4;
		String gradient = null;
		switch(pos) {
			case 0:
				gradient = "0";
				break;
			case 1:
				gradient = "1";
				break;
			case 2:
				gradient = INFINITY;
				break;
			case 3:
				gradient = "-1";
		}
		return gradient;
	}
	private T getBoundValue(String val,String select) throws Exception {
		byte pos;
		if(val.compareTo(INFINITY)==0)
			pos = 2;
		else {
			try {
				byte value = Byte.parseByte(val);
				switch(value) {
					case 0:
						pos = 0;
						break;
					case 1:
						pos = 1;
						break;
					case -1:
						pos = 3;
						break;
					default:
						throw new Exception();
				}
			} catch(Exception e) {
				String msg = "The initial gradient of a boundary\n"
				        +"point must be one of -1,0,1 or "+INFINITY;
				throw new Exception(msg);
			}
		}
		if(select.compareTo(BND_VAL_POINT[1])==0)
			pos += 4;
		return math.util.cast(pos);
	}
	@Override
	int initGui(GridBagLayout gb,GridBagConstraints c,int row) {
		c.fill = GridBagConstraints.BOTH;
		c.gridx = 1;
		c.gridy = row;
		c.weightx = 3;
		c.weighty = 1;
		c.gridwidth = 1;
		c.gridheight = 5;
		TitledBorder tb = BorderFactory.createTitledBorder("Sequence");
		buildTable();
		seqTableScroller = new JScrollPane(seqTable);
		int policy = ScrollPaneConstants.VERTICAL_SCROLLBAR_AS_NEEDED;
		seqTableScroller.setVerticalScrollBarPolicy(policy);
		policy = ScrollPaneConstants.HORIZONTAL_SCROLLBAR_NEVER;
		seqTableScroller.setHorizontalScrollBarPolicy(policy);
		seqTableScroller.setBorder(tb);
		gb.setConstraints(seqTableScroller,c);
		add(seqTableScroller);
		c.fill = GridBagConstraints.BOTH;
		c.gridx = 0;
		c.gridy = row;
		c.weightx = 1;
		c.weighty = 0;
		c.gridwidth = 1;
		c.gridheight = 1;
		tb = BorderFactory.createTitledBorder("1. Boundary Value Type");
		startBTComboBox = new JComboBox(BND_VAL_TYPES);
		startBTComboBox.setBorder(tb);
		gb.setConstraints(startBTComboBox,c);
		add(startBTComboBox);
		c.fill = GridBagConstraints.BOTH;
		c.gridx = 0;
		c.gridy = ++row;
		c.weightx = 1;
		c.weighty = 0;
		c.gridwidth = 1;
		c.gridheight = 1;
		tb = BorderFactory.createTitledBorder("1. Boundary Value");
		startBVComboBox = new JComboBox(BND_VAL_POINT);
		startBVComboBox.setBorder(tb);
		gb.setConstraints(startBVComboBox,c);
		add(startBVComboBox);
		c.fill = GridBagConstraints.BOTH;
		c.gridx = 0;
		c.gridy = ++row;
		c.weightx = 1;
		c.weighty = 0;
		c.gridwidth = 1;
		c.gridheight = 1;
		tb = BorderFactory.createTitledBorder("2. Boundary Value Type");
		endBTComboBox = new JComboBox(BND_VAL_TYPES);
		endBTComboBox.setBorder(tb);
		gb.setConstraints(endBTComboBox,c);
		add(endBTComboBox);
		c.fill = GridBagConstraints.BOTH;
		c.gridx = 0;
		c.gridy = ++row;
		c.weightx = 1;
		c.weighty = 0;
		c.gridwidth = 1;
		c.gridheight = 1;
		tb = BorderFactory.createTitledBorder("2. Boundary Value");
		endBVComboBox = new JComboBox(BND_VAL_POINT);
		endBVComboBox.setBorder(tb);
		gb.setConstraints(endBVComboBox,c);
		add(endBVComboBox);
		c.fill = GridBagConstraints.BOTH;
		c.gridx = 0;
		c.gridy = ++row;
		c.weightx = 1;
		c.weighty = 1;
		c.gridwidth = 2;
		c.gridheight = 1;

		JPanel p = new JPanel();
		gb.setConstraints(p,c);
		add(p);
		ItemListener l = new AdjustBoundValuesListener(startBTComboBox,startBVComboBox);
		startBTComboBox.addItemListener(l);
		l = new AdjustBoundValuesListener(endBTComboBox,endBVComboBox);
		endBTComboBox.addItemListener(l);
		return ++row;
	}
	private void buildTable() {
		seqTableModel = new SeqenceTableModel(new String[]{"","Cell","Gradient"});
		seqTable = new JTable(seqTableModel);
		seqTable.setRowSelectionAllowed(true);
		seqTable.setColumnSelectionAllowed(false);
		CellEditorListener autoInserter = new CellEditorListener() {

			@Override
			public void editingCanceled(ChangeEvent arg0) {}
			@Override
			public void editingStopped(ChangeEvent e) {
				TableCellEditor editor = (TableCellEditor)e.getSource();
				String val = (String)editor.getCellEditorValue();
				int rIdx = seqTable.getEditingRow();
				int cIdx = seqTable.getEditingColumn();
				seqTableModel.setValueAt(val,rIdx,cIdx);
			}
		};
		StringEditor se = new StringEditor(true);
		ColorRenderer cr = new ColorRenderer();
		se.addCellEditorListener(autoInserter);
		TableColumn col = seqTable.getColumnModel().getColumn(0);
		col.setMaxWidth(30);
		col.setMinWidth(30);
		col = seqTable.getColumnModel().getColumn(1);
		col.setCellEditor(se);
		col.setCellRenderer(cr);
		col = seqTable.getColumnModel().getColumn(2);
		col.setCellEditor(se);
		col.setCellRenderer(cr);
		JPopupMenu popupMenu = new JPopupMenu();
		JMenuItem insertRowItem = new JMenuItem("Insert");
		insertRowItem.addMouseListener(new MouseAdapter() {

			@Override
			public void mouseReleased(MouseEvent me) {
				stopEditing(seqTable);
				int[] selRows = seqTable.getSelectedRows();
				seqTableModel.insertRow(selRows[0]);
			}
		});
		JMenuItem deleteRowItem = new JMenuItem("Delete");
		deleteRowItem.addMouseListener(new MouseAdapter() {

			@Override
			public void mouseReleased(MouseEvent me) {
				stopEditing(seqTable);
				int[] selRows = seqTable.getSelectedRows();
				seqTableModel.removeRows(selRows);
			}
		});
		final JMenuItem copyItem = new JMenuItem("Copy");
		copyItem.addMouseListener(new MouseAdapter() {

			@Override
			public void mouseReleased(MouseEvent me) {
				stopEditing(seqTable);
				seqTableModel.bufferRows(seqTable.getSelectedRows());
			}
		});
		final JMenuItem pasteItem = new JMenuItem("Paste");
		pasteItem.addMouseListener(new MouseAdapter() {

			@Override
			public void mouseReleased(MouseEvent me) {
				stopEditing(seqTable);
				int[] selRows = seqTable.getSelectedRows();
				seqTableModel.copyBuffer(selRows[0]);
			}
		});
		final JMenuItem incXItem = new JMenuItem("x -> x+1");
		incXItem.addMouseListener(new ReselectMouseListener() {

			@Override
			void execute() {
				seqTableModel.incX(seqTable.getSelectedRows());
			}
		});
		final JMenuItem decXItem = new JMenuItem("x -> x-1");
		decXItem.addMouseListener(new ReselectMouseListener() {

			@Override
			void execute() {
				seqTableModel.decX(seqTable.getSelectedRows());
			}
		});
		final JMenuItem incYItem = new JMenuItem("y -> y+1");
		incYItem.addMouseListener(new ReselectMouseListener() {

			@Override
			void execute() {
				seqTableModel.incY(seqTable.getSelectedRows());
			}
		});
		final JMenuItem decYItem = new JMenuItem("y -> y-1");
		decYItem.addMouseListener(new ReselectMouseListener() {

			@Override
			void execute() {
				seqTableModel.decY(seqTable.getSelectedRows());
			}
		});
		popupMenu.add(insertRowItem);
		popupMenu.add(deleteRowItem);
		popupMenu.add(copyItem);
		popupMenu.add(pasteItem);
		popupMenu.add(incXItem);
		popupMenu.add(decXItem);
		popupMenu.add(incYItem);
		popupMenu.add(decYItem);
		seqTable.setComponentPopupMenu(popupMenu);
		ListSelectionModel selModel = seqTable.getSelectionModel();
		selModel.addListSelectionListener(new ListSelectionListener() {

			@Override
			public void valueChanged(ListSelectionEvent arg0) {
				boolean b = seqTable.getSelectedRows().length!=0;
				enablePopupMenues(b);
				clearEstimation();
			}
		});
		selModel.setSelectionMode(ListSelectionModel.SINGLE_INTERVAL_SELECTION);
		enablePopupMenues(false);
	}
	private void enablePopupMenues(boolean b) {
		JPopupMenu pm = seqTable.getComponentPopupMenu();
		for(int i = 0;i<pm.getComponentCount();i++)
			pm.getComponent(i).setEnabled(b);
	}
	@Override
	public void setEnabled(boolean b) {
		super.setEnabled(b);
		startBTComboBox.setEnabled(b);
		startBVComboBox.setEnabled(b);
		endBTComboBox.setEnabled(b);
		endBVComboBox.setEnabled(b);
		seqTable.setEnabled(b);
		seqTableScroller.setEnabled(b);
	}
	@Override
	public void clear() {
		super.clear();
		stopEditing(seqTable);
		seqTableModel.clear();
		startBTComboBox.setSelectedItem(null);
		startBVComboBox.setSelectedItem(null);
		endBTComboBox.setSelectedItem(null);
		endBVComboBox.setSelectedItem(null);
	}
	private static void stopEditing(JTable table) {
		int rIdx = table.getEditingRow();
		int cIdx = table.getEditingColumn();
		if(rIdx!=-1&&cIdx!=-1)
			table.getCellEditor(rIdx,cIdx).stopCellEditing();
	}
	private class StringEditor extends DefaultCellEditor {

		public StringEditor(boolean editable) {
			super(new JTextField());
			setClickCountToStart(1);
			((JTextField)getComponent()).setEditable(editable);
			((JTextField)getComponent()).setBackground(Globals.INPUT_BG);
		}
	}
	private class ColorRenderer extends DefaultTableCellRenderer {

		@Override
		public Component getTableCellRendererComponent(JTable t,Object v, boolean sel, 
				boolean b, int row,int col) {
			// TODO Auto-generated method stub
			Component c = super.getTableCellRendererComponent(t,v,sel,b,row,col);
			c.setBackground(Globals.INPUT_BG);
			return c;
		}
	}
	private abstract class ReselectMouseListener extends MouseAdapter {

		@Override
		public void mouseReleased(MouseEvent me) {
			stopEditing(seqTable);
			final int i0 = seqTable.getSelectionModel().getMinSelectionIndex();
			final int i1 = seqTable.getSelectionModel().getMaxSelectionIndex();
			execute();
			SwingUtilities.invokeLater(new Runnable() {

				public void run() {
					seqTable.getSelectionModel().addSelectionInterval(i0,i1);
				}
			});
		}
		abstract void execute();
	}
	private class AdjustBoundValuesListener implements ItemListener {

		private JComboBox btComboBox;
		private JComboBox bvComboBox;
		AdjustBoundValuesListener(JComboBox btCBox,JComboBox bvCBox) {
			this.btComboBox = btCBox;
			this.bvComboBox = bvCBox;
		}
		@Override
		public void itemStateChanged(ItemEvent ie) {
			bvComboBox.removeAllItems();
			String si = (String)btComboBox.getSelectedItem();
			if(si!=null) {
				String[] items = BND_VAL_POINT;
				if(si.compareTo(BND_VAL_TYPES[0])!=0)
					items = BND_VAL_GRAD;
				for(String item:items)
					bvComboBox.addItem(item);
			}
			clearEstimation();
		}
	}
}
