/**
	Copyright (c) 2009-2013 by Zoltan Palmer

	This file is part of FroZenLight.

    FroZenLight is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    FroZenLight is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with FroZenLight.  If not, see http://www.gnu.org/licenses/
*/
package pattern;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;

import math.VecMath;

import common.Globals;

public class LightCharPattern<T extends Number> extends BaseLightPattern<T> {

	public final static byte MAX_RADIUS = 70;
	private static enum DIR {
		LEFT,RIGHT,DOWN
	}
	private static final byte ORI_HB = 0;
	private static final byte ORI_HT = 1;
	private static final byte ORI_DP = 2;
	private static final byte ORI_DN = 3;
	private static final byte ORI_VP = 4;
	private static final byte ORI_VN = 5;
	private LightChar[] lightChars;
	private String text;
	// String letterPos = "";
	public LightCharPattern(String name,String txt,LightChar[] lcs,VecMath<T> mt) throws Exception {
		super(mt);
		this.lightChars = lcs;
		this.text = txt;
		setName(name);
		setText(txt);
		short maxPrec = Globals.PRECISIONS[Globals.PRECISIONS.length-1];
		T maxR = mt.util.cast(MAX_RADIUS);
		setMaxShapeSize(mt.util.copy(maxR,maxPrec));
		setStartBndValType(GRAD_LEFT);
		setStartIndex(0);
		setPropagating(true);
	}
	public LightCharPattern(String name,String txt,boolean store,LightChar[] lcs,VecMath<T> mt)
	        throws Exception {
		this(name,txt,lcs,mt);
		setStoreCalculation(store);
	}
	public LightCharPattern(LightCharPattern<T> p) {
		super(p);
		this.lightChars = p.lightChars;
		this.text = p.text;
	}
	@Override
	protected int[] calcDigitLossLowHigh() {
		float[] lossLowHigh = calcSequenceDigitLossLowHigh();
		return new int[]{(int)lossLowHigh[0],(int)(lossLowHigh[1]+0.5)};
	}
	public String getText() {
		return text;
	}
	public void setText(String text) throws Exception {
		this.text = text;
		if(text!=null) {
			String rText = reduce(text);
			gridSize = calcGridSize(rText);
			buildPattern(rText);
			setEndBndValType(gridSize%2==1?GRAD_RIGHT:GRAD_LEFT);
		}
	}
	public void setParameters(T[] gridCorner,T gap,T r) {
		// Log.getIstc().logln(letterPos);
		super.setParameters(gridCorner,gap,r);
		T t = mt.div(r,gap);
		T m = mt.div(t,mt.sqrt(mt.sub(mt.ONE,mt.mul(mt.TWO,t))));
		setStartBndVal(m);
		setEndBndVal(gridSize%2==1?mt.neg(m):m);
	}
	private void buildPattern(String txt) throws Exception {
		String charList = listUnknownChars(txt);
		if(charList==null) {
			m_init = new ArrayList<Float>();
			cells = new ArrayList<short[]>();
			txt = prepare(txt,gridSize);
			int charCtr = 0;
			boolean flow = false;
			for(int row = gridSize-2;row>=0;row--) {
				for(int col = flow?0:gridSize-2;col>=0&&col<gridSize-1;col += flow?1:-1) {
					if(col==gridSize-2) {
						LightChar lc = getLightChar("eol").copy();
						lc.move((short)col,(short)row);
						append(lc,flow?DIR.RIGHT:DIR.DOWN);
					} else if(col==0) {
						LightChar lc = getLightChar("bol").copy();
						lc.move((short)col,(short)row);
						append(lc,flow?DIR.DOWN:DIR.LEFT);
					} else {
						LightChar lc = getLightChar(txt.charAt(charCtr++)+"").copy();
						lc.move((short)col,(short)row);
						// letterPos += "Circle "+(cells.size()+1)+" :"+lc.getChar()+"\n";
						append(lc,flow?DIR.RIGHT:DIR.LEFT);
					}
				}
				flow = !flow;
			}
		} else {
			String chars = "";
			for(LightChar lc:lightChars)
				if(lc.getChar().compareTo("eol")!=0&&lc.getChar().compareTo("bol")!=0)
					chars += lc.getChar();
			throw new Exception("Found unsupported characters: "+charList+"\nAvailable: "+chars);
		}
		m_init.set(0,0f);
		m_init.set(m_init.size()-1,0f);
	}
	private void append(LightChar s,DIR dir) {
		if(cells.size()>0) {
			if(dir==DIR.DOWN)
				mergeByRemove(s.m_init,0f);
			else {
				boolean reverse = dir==DIR.LEFT;
				byte o1 = orientation(cells,true);
				int sign = reverse?-1:1;
				if(reverse&&s.getChar().compareToIgnoreCase("bol")!=0
				        &&s.getChar().compareToIgnoreCase("eol")!=0) {
					Collections.reverse(s.cells);
					Collections.reverse(s.m_init);
				}
				byte o2 = orientation(s.cells,false);
				switch(o1) {
					case ORI_HB:
						switch(o2) {
							case ORI_HB:
								mergeByInsert(false,100f);
								break;
							case ORI_DP:
								mergeByRemove(s.m_init,-2.8f*sign);
								break;
							case ORI_VP:
								mergeByRemove(s.m_init,-1.23f*sign);
						}
						break;
					case ORI_HT:
						switch(o2) {
							case ORI_HT:
								mergeByInsert(true,100f);
								break;
							case ORI_DN:
								mergeByRemove(s.m_init,2.8f*sign);
								break;
							case ORI_VN:
								mergeByRemove(s.m_init,1.23f*sign);
						}
						break;
					case ORI_DP:
						switch(o2) {
							case ORI_HT:
								mergeByRemove(s.m_init,-2.8f*sign);
								break;
							case ORI_DN:
								mergeByRemove(s.m_init,100f);
								break;
							case ORI_VN:
								mergeByRemove(s.m_init,2.7f*sign);
						}
						break;
					case ORI_DN:
						switch(o2) {
							case ORI_HB:
								mergeByRemove(s.m_init,2.8f*sign);
								break;
							case ORI_DP:
								mergeByRemove(s.m_init,100f);
								break;
							case ORI_VP:
								mergeByRemove(s.m_init,-2.7f*sign);
								break;
						}
						break;
					case ORI_VP:
						switch(o2) {
							case ORI_HT:
								mergeByRemove(s.m_init,-1.23f*sign);
								break;
							case ORI_DN:
								mergeByRemove(s.m_init,-2.7f*sign);
								break;
							case ORI_VN:
								mergeByRemove(s.m_init,100f);
						}
						break;
					case ORI_VN:
						switch(o2) {
							case ORI_HB:
								mergeByRemove(s.m_init,1.23f*sign);
								break;
							case ORI_DP:
								mergeByRemove(s.m_init,2.7f*sign);
								break;
							case ORI_VP:
								mergeByRemove(s.m_init,100f);
						}
				}
			}
		}
		cells.addAll(s.cells);
		m_init.addAll(s.m_init);
	}
	private void mergeByRemove(List<Float> l3,float m) {
		m_init.remove(m_init.size()-1);
		cells.remove(cells.size()-1);
		l3.set(0,m);
	}
	private void mergeByInsert(boolean botOrTop,float m) {
		short[] cell = cells.get(cells.size()-1).clone();
		cell[1] += botOrTop?-1:1;
		cells.add(cell);
		m_init.add(m);
	}
	private static byte orientation(List<short[]> cellList,boolean frontOrEnd) {
		int i0 = frontOrEnd?cellList.size()-1:0;
		int inc = frontOrEnd?-1:1;
		short x1 = cellList.get(i0)[0];
		short y1 = cellList.get(i0)[1];
		short x2 = cellList.get(i0+inc)[0];
		short y2 = cellList.get(i0+inc)[1];
		if(y1!=y2) {
			if(x1!=x2)
				return y1>y2==frontOrEnd?ORI_DP:ORI_DN;
			return y1>y2==frontOrEnd?ORI_VP:ORI_VN;
		}
		for(int i = i0+2*inc;y1==(y2 = cellList.get(i)[1]);i += inc)
			;
		return y1>y2?ORI_HT:ORI_HB;
	}
	private LightChar getLightChar(String sbl) {
		for(LightChar lc:lightChars)
			if(lc.getChar().compareToIgnoreCase(sbl)==0)
				return lc;
		return null;
	}
	private String listUnknownChars(String txt) {
		StringBuffer sb = new StringBuffer();
		for(char c:txt.toCharArray())
			if(getLightChar(c+"")==null)
				sb.append(c).append(" ");
		if(sb.length()>0) {
			sb.deleteCharAt(sb.length()-1);
			return sb.toString();
		}
		return null;
	}
	public static String reduce(String text) {
		text = text.trim();
		text = text.replace("\n","");
		text = text.replace('\t',' ');
		text = text.replace('\b',' ');
		text = text.replace('\f',' ');
		text = text.replace('\r',' ');
		text = text.replace('`','\'');
		// while(text.indexOf("  ") != -1)
		// text = text.replace("  "," ");
		return text;
	}
	private static String prepare(String text,short size) {
		StringBuffer tbuf = new StringBuffer();
		for(int i = 0;i<size-1;i++) {
			boolean e = i%2==0;
			for(int j = e?size-4:0;j>=0&&j<size-3;j += e?-1:+1) {
				int idx = i*(size-3)+j;
				tbuf.append(idx<text.length()?text.charAt(idx):' ');
			}
		}
		return tbuf.toString();
	}
	public static String[] preview(String text,short size) {
		text = text.replace(" ","/");
		if(size>0) {
			String[] lines = new String[size-1];
			String space = "";
			for(int i = 0;i<size-3;i++)
				space += "/";
			for(int i = 0;i<lines.length;i++) {
				int k0 = i*(size-3);
				int k1 = Math.min(k0+size-3,text.length());
				if(k0<text.length()) {
					int k2 = k1%(size-3);
					String insert = k2!=0?space.substring(k2):"";
					lines[i] = "["+text.substring(k0,k1)+insert+"]";
				} else
					lines[i] = "["+space+"]";
			}
			return lines;
		}
		return new String[0];
	}
	public static short calcGridSize(String text) {
		return (short)Math.ceil(2+Math.sqrt(1+text.length()));
	}
	public String toString() {
		StringBuffer sb = new StringBuffer();
		sb.append("text = ");
		sb.append(getText());
		sb.append("\n");
		sb.append("store_calc = ");
		sb.append(isStoreCalcEnabled());
		return sb.toString();
	}
}
