/**
	Copyright (c) 2009-2013 by Zoltan Palmer

	This file is part of FroZenLight.

    FroZenLight is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    FroZenLight is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with FroZenLight.  If not, see http://www.gnu.org/licenses/
*/
package pattern;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;

import lattice.Circle;
import math.IFunction;
import math.Util;
import math.VecMath;

import common.Globals;
import common.Log;
import common.Profiler;

final public class LightPatternSolver<T extends Number> {

	private final static String[] HEADERS;
	public final static short INI_AC = 20;
	private final static float[] OPT_DAMPS = new float[11];
	static {
		OPT_DAMPS[0] = 0.9f;
		OPT_DAMPS[1] = 0.9f;
		OPT_DAMPS[2] = 0.9f;
		OPT_DAMPS[3] = 0.86f;
		OPT_DAMPS[4] = 0.8f;
		OPT_DAMPS[5] = 0.73f;
		OPT_DAMPS[6] = 0.65f;
		OPT_DAMPS[7] = 0.56f;
		OPT_DAMPS[8] = 0.46f;
		OPT_DAMPS[9] = 0.35f;
		OPT_DAMPS[10] = 0.23f;
		HEADERS = new String[4];
		HEADERS[0] = "Circle  ";
		HEADERS[1] = "Accuracy  ";
		HEADERS[2] = "Gradient (Truncated)               ";
		HEADERS[3] = "Delta ";
	}
	private final VecMath<T> mt;
	private T opt_damp;
	private T damp;
	private int accuracy;
	private int updateAc;
	private final int margin;
	private ArrayList<IProgressListener> listeners;
	private long updateTime;
	private boolean term;
	private long startTime;
	private List<Node> curNodes;
	private List<Node> newNodes;
	private int curAc;
	private boolean log;
	private int threads;
	public LightPatternSolver(float rOverGap,boolean log,VecMath<T> vm) {
		this.log = log;
		this.mt = vm;
		this.margin = 30;
		float p = 2*rOverGap;
		float p0 = (int)(p*10)/10f;
		float p1 = ((int)(p*10)+1)/10f;
		float d0 = OPT_DAMPS[(int)(p*10)];
		float d1 = OPT_DAMPS[Math.min((int)(p*10)+1,10)];
		float t = (p-p0)/(p1-p0);
		setInitialDamp(mt.util.round(mt.util.cast(OPT_DAMPS[9]),2));
		setOptimalDamp(mt.util.round(mt.util.cast(d0+t*(d1-d0)),2));
		threads = Runtime.getRuntime().availableProcessors();
	}
	private void init(List<Circle<T>> circles,List<Float> m_init) {
		startTime = System.currentTimeMillis();
		correctGradients(circles,m_init);
		curNodes = Collections.synchronizedList(new ArrayList<Node>());
		newNodes = Collections.synchronizedList(new ArrayList<Node>());
		for(int k = 0;k < circles.size();k++)
			newNodes.add(null);
		updateAc = 0;
		curAc = 0;
		for(int i = 0;i<circles.size();i++) {
			Node n = new Node();
			n.m0 = mt.util.copy(mt.ZERO,10);
			if(m_init!=null)
				n.m0 = mt.util.cast(m_init.get(i));
			n.m = n.m0;
			curNodes.add(n);
		}
	}
	/**
	/* Calculate pattern. Boundary condition is given by two gradients. Returns
	/* gradients of lines connecting circle centers and refl. points.
	*/
	public List<T[]> solve(T[] mBnd,boolean[] dirBnd,List<Circle<T>> circles,List<Float> m_init) {
		init(circles,m_init);
		T[] moveVecStart = null;
		T[] moveVecEnd = null;
		for(int i = 0;i<circles.size();i++) {
			Node n = curNodes.get(i);
			T[] ctr1;
			if(i==0) {
				moveVecStart = calcMoveVector(circles.get(0),mBnd[0],!dirBnd[0]);
				ctr1 = circles.get(0).getCenter().clone();
				mt.add(ctr1,moveVecStart,ctr1);
			} else
				ctr1 = circles.get(i-1).getCenter();
			T[] ctr2;
			if(i==circles.size()-1) {
				moveVecEnd = calcMoveVector(circles.get(i),mBnd[1],dirBnd[1]);
				ctr2 = circles.get(i).getCenter().clone();
				mt.add(ctr2,moveVecEnd,ctr2);
			} else
				ctr2 = circles.get(i+1).getCenter();
			updatePoint(n,circles.get(i),ctr1,ctr2);
		}
		for(int i = 0;i<circles.size();i++) {
			Node n = curNodes.get(i);
			if(i==0) {
				n.dy1 = mBnd[0];
				n.dx1 = mt.ONE;
			} else {
				T[] pt1 = curNodes.get(i-1).reflPnt;
				n.dy1 = mt.sub(n.reflPnt[1],pt1[1]);
				n.dx1 = mt.sub(n.reflPnt[0],pt1[0]);
			}
			if(i<circles.size()-1) {
				T[] pt2 = curNodes.get(i+1).reflPnt;
				n.dy2 = mt.sub(n.reflPnt[1],pt2[1]);
				n.dx2 = mt.sub(n.reflPnt[0],pt2[0]);
			} else {
				n.dy2 = mBnd[1];
				n.dx2 = mt.ONE;
			}
		}
		List<SolveRbl> rbls = new ArrayList<SolveRbl>();
		for(int i = 0;i<Math.min(circles.size(),threads);i++)
			rbls.add(new SolveGradGradRbl(circles,mBnd,moveVecStart,moveVecEnd));
		return solve(rbls);
	}
	private List<T[]> solve(List<SolveRbl> rbls){
		if(Globals.LOG_VERBOSE)
			printReflectionNodes(curNodes,(byte)0);
		Common com = new Common(curNodes.size()-1);
		for(int i = 0;i<rbls.size();i++){
			rbls.get(i).com = com;
			if(i>0)
				new Thread(rbls.get(i)).start();
		}
		int steps = 0;
		while(curAc<accuracy&&!term){
			for(int k = 0;k < newNodes.size();k++)
				newNodes.set(k,null);
			com.reset();
			for(int i = com.nextPos();i!=-1&&!term;i = com.nextPos())
				rbls.get(0).updateNode(i);
			for(int k = 0;k<newNodes.size()&&!term;k++)
				if(newNodes.get(k)==null)
					k = -1;
			for(int k = 0;k<newNodes.size();k++)
				curNodes.set(k,newNodes.get(k));
			if(curAc>INI_AC&&mt.cmp(damp,opt_damp)!=0)
				damp = opt_damp;
			handleProgressListeners(curAc);
			steps++;
		}
		if(!term) {
			printInfo(steps,startTime,curNodes,(byte)0);
			return getReflPoints(curNodes);
		} 
		com.reset();
		return null;
	}
	private class Common {
		final int maxPos;
		int pos = 0;
		boolean active = false;
		Common(int mp){this.maxPos = mp;}
		synchronized int nextPosOrWait(){
			while(!active)
				try{
					wait();
				} catch(InterruptedException ie){}
			if(pos>maxPos){
				try{
					wait();
				} catch(InterruptedException ie){}
				return nextPosOrWait();
			}
			return ++pos-1;
		}
		synchronized int nextPos() {return pos>maxPos?-1:++pos-1;}
		synchronized void reset(){
			pos = 0;
			active = true;
			notifyAll();
		}
	}
	private abstract class SolveRbl implements Runnable {
		List<Circle<T>> circles;
		Common com;
		T[] pt;
		T[] pt1;
		T[] pt2;
		public void run(){
			while(!term)
				updateNode(com.nextPosOrWait());
		}
		final void updateNodeGradient1(int i,Node newNode){
			if(threads==1){
				newNode.dy1 = newNodes.get(i-1).dy2;
				newNode.dx1 = newNodes.get(i-1).dx2;
			} else {
				newNode.dy1 = mt.sub(pt1[1],pt[1]);
				newNode.dx1 = mt.sub(pt1[0],pt[0]);
			}
		}
		final void updateNode(int i){
			Node curNode = curNodes.get(i);
			Node newNode = new Node(curNode);
			pt = curNode.reflPnt;
			prepareUpdateNode(i,newNode,curNode);
			calcNewNode(newNode,curNode,circles.get(i),curAc,pt1,pt2);
			curAc = Math.max(curAc,newNode.ac);
			releaseMemory(i);
			newNodes.set(i,newNode);
		}
		abstract void prepareUpdateNode(int i,Node neew,Node cur);
		final void releaseMemory(int i){
			// A few nodes might be left undeleted.
			if(--i>0&&newNodes.get(i)!=null)
				while(i>0&&newNodes.get(i-1)!=null)
					curNodes.set(i--,null);
		}
	}
	private final class SolveGradGradRbl extends SolveRbl {
		private T[] moveVecStart;
		private T[] moveVecEnd;
		private T[] mBnd;
		SolveGradGradRbl(List<Circle<T>> cls,T[] mbd,T[] v1,T[] v2){
			this.circles = cls; 
			this.mBnd = mbd;
			this.moveVecStart = v1; 
			this.moveVecEnd = v2;
		}
		void prepareUpdateNode(int i,Node newNode,Node curNode){
			if(i==0) {
				newNode.dy1 = mBnd[0];
				newNode.dx1 = mt.ONE;
				// Create virtual point in direction of the bound ray
				pt1 = curNodes.get(0).reflPnt.clone();
				mt.add(pt1,moveVecStart,pt1);
			} else {
				pt1 = curNodes.get(i-1).reflPnt;
				updateNodeGradient1(i,newNode);
			}
			if(i<circles.size()-1) {
				pt2 = curNodes.get(i+1).reflPnt;
				newNode.dy2 = mt.sub(pt[1],pt2[1]);
				newNode.dx2 = mt.sub(pt[0],pt2[0]);
			} else {
				newNode.dy2 = mBnd[1];
				newNode.dx2 = mt.ONE;
				// Create virtual point in direction of the bound ray
				pt2 = curNodes.get(i).reflPnt.clone();
				mt.add(pt2,moveVecEnd,pt2);
			}
		}
	}
	public List<T[]> solve(T[] start,T endM,boolean endDir,List<Circle<T>> circles,
	        List<Float> m_init) {
		init(circles,m_init);
		T[] moveVec = null;
		for(int i = 0;i<circles.size();i++) {
			Node n = curNodes.get(i);
			T[] ctr1 = i==0?start:circles.get(i-1).getCenter();
			T[] ctr2;
			if(i==circles.size()-1) {
				moveVec = calcMoveVector(circles.get(i),endM,endDir);
				ctr2 = circles.get(i).getCenter().clone();
				mt.add(ctr2,moveVec,ctr2);
			} else
				ctr2 = circles.get(i+1).getCenter();
			updatePoint(n,circles.get(i),ctr1,ctr2);
		}
		for(int i = 0;i<circles.size();i++) {
			Node n = curNodes.get(i);
			T[] pt1 = i==0?start:curNodes.get(i-1).reflPnt;
			n.dy1 = mt.sub(n.reflPnt[1],pt1[1]);
			n.dx1 = mt.sub(n.reflPnt[0],pt1[0]);
			if(i<circles.size()-1) {
				T[] pt2 = curNodes.get(i+1).reflPnt;
				n.dy2 = mt.sub(n.reflPnt[1],pt2[1]);
				n.dx2 = mt.sub(n.reflPnt[0],pt2[0]);
			} else {
				n.dy2 = endM;
				n.dx2 = mt.ONE;
			}
		}
		List<SolveRbl> rbls = new ArrayList<SolveRbl>();
		for(int i = 0;i<Math.min(circles.size(),threads);i++)
			rbls.add(new SolvePointGradRbl(circles,start,moveVec,endM));
		return solve(rbls);
	}
	private final class SolvePointGradRbl extends SolveRbl {
		private T[] start;
		private T[] moveVec;
		private T endM;
		SolvePointGradRbl(List<Circle<T>> cls,T[] s,T[] v,T em){
			this.circles = cls;
			this.moveVec = v;  
			this.start = s;
			this.endM = em;
		}
		void prepareUpdateNode(int i,Node newNode,Node curNode){
			if(i==0) {
				newNode.dy1 = mt.sub(pt[1],start[1]);
				newNode.dx1 = mt.sub(pt[0],start[0]);
				pt1 = start;
			} else {
				pt1 = curNodes.get(i-1).reflPnt;
				updateNodeGradient1(i,newNode);
			}
			if(i<circles.size()-1) {
				pt2 = curNodes.get(i+1).reflPnt;
				newNode.dy2 = mt.sub(pt[1],pt2[1]);
				newNode.dx2 = mt.sub(pt[0],pt2[0]);
			} else {
				newNode.dy2 = endM;
				newNode.dx2 = mt.ONE;
				// Create virtual point in direction of the bound ray
				pt2 = curNodes.get(i).reflPnt.clone();
				mt.add(pt2,moveVec,pt2);
			}
		}
	}
	public List<T[]> solve(T stM,boolean stDir,T[] end,List<Circle<T>> circles,List<Float> m_init) {
		init(circles,m_init);
		T[] moveVec = null;
		for(int i = 0;i<circles.size();i++) {
			Node n = curNodes.get(i);
			T[] ctr1;
			if(i==0) {
				moveVec = calcMoveVector(circles.get(0),stM,!stDir);
				ctr1 = circles.get(0).getCenter().clone();
				mt.add(ctr1,moveVec,ctr1);
			} else
				ctr1 = circles.get(i-1).getCenter();
			T[] ctr2 = i==circles.size()-1?end:circles.get(i+1).getCenter();
			updatePoint(n,circles.get(i),ctr1,ctr2);
		}
		for(int i = 0;i<circles.size();i++) {
			Node n = curNodes.get(i);
			T[] pt1;
			if(i==0) {
				n.dy1 = stM;
				n.dx1 = mt.ONE;
			} else {
				pt1 = curNodes.get(i-1).reflPnt;
				n.dy1 = mt.sub(n.reflPnt[1],pt1[1]);
				n.dx1 = mt.sub(n.reflPnt[0],pt1[0]);
			}
			T[] pt2 = i==circles.size()-1?end:curNodes.get(i+1).reflPnt;
			n.dy2 = mt.sub(n.reflPnt[1],pt2[1]);
			n.dx2 = mt.sub(n.reflPnt[0],pt2[0]);
		}
		List<SolveRbl> rbls = new ArrayList<SolveRbl>();
		for(int i = 0;i<Math.min(circles.size(),threads);i++)
			rbls.add(new SolveGradPointRbl(circles,end,moveVec,stM));
		return solve(rbls);
	}
	private final class SolveGradPointRbl extends SolveRbl {
		private T[] end;
		private T[] moveVec;
		private T stM;
		SolveGradPointRbl(List<Circle<T>> cls,T[] e,T[] v,T sm){
			this.circles = cls;
			this.moveVec = v;  
			this.end = e;
			this.stM = sm;
		}
		void prepareUpdateNode(int i,Node newNode,Node curNode){
			if(i==0) {
				newNode.dy1 = stM;
				newNode.dx1 = mt.ONE;
				// Create virtual point in direction of the bound ray
				pt1 = curNodes.get(0).reflPnt.clone();
				mt.add(pt1,moveVec,pt1);
			} else {
				pt1 = curNodes.get(i-1).reflPnt;
				updateNodeGradient1(i,newNode);
			}
			pt2 = i==circles.size()-1?end:curNodes.get(i+1).reflPnt;
			newNode.dy2 = mt.sub(pt[1],pt2[1]);
			newNode.dx2 = mt.sub(pt[0],pt2[0]);
		}
	}
	public List<T[]> solve(T[] start,T[] end,List<Circle<T>> circles,List<Float> m_init) {
		init(circles,m_init);
		for(int i = 0;i<circles.size();i++) {
			Node n = curNodes.get(i);
			T[] ctr1 = i==0?start:circles.get(i-1).getCenter();
			T[] ctr2 = i==circles.size()-1?end:circles.get(i+1).getCenter();
			updatePoint(n,circles.get(i),ctr1,ctr2);
		}
		for(int i = 0;i<circles.size();i++) {
			Node n = curNodes.get(i);
			T[] pt1 = i==0?start:curNodes.get(i-1).reflPnt;
			T[] pt2 = i==circles.size()-1?end:curNodes.get(i+1).reflPnt;
			n.dy1 = mt.sub(n.reflPnt[1],pt1[1]);
			n.dx1 = mt.sub(n.reflPnt[0],pt1[0]);
			n.dy2 = mt.sub(n.reflPnt[1],pt2[1]);
			n.dx2 = mt.sub(n.reflPnt[0],pt2[0]);
		}
		List<SolveRbl> rbls = new ArrayList<SolveRbl>();
		for(int i = 0;i<Math.min(circles.size(),threads);i++)
			rbls.add(new SolvePointPointRbl(circles,start,end));
		return solve(rbls);
	}
	private final class SolvePointPointRbl extends SolveRbl {
		private T[] start;
		private T[] end;
		SolvePointPointRbl(List<Circle<T>> cls,T[] s,T[] e){
			this.circles = cls;
			this.start = s;  
			this.end = e;
		}
		void prepareUpdateNode(int i,Node newNode,Node curNode){
			if(i==0) {
				newNode.dy1 = mt.sub(pt[1],start[1]);
				newNode.dx1 = mt.sub(pt[0],start[0]);
				pt1 = start;
			} else {
				pt1 = curNodes.get(i-1).reflPnt;
				updateNodeGradient1(i,newNode);
			}
			pt2 = i==circles.size()-1?end:curNodes.get(i+1).reflPnt;
			newNode.dy2 = mt.sub(pt[1],pt2[1]);
			newNode.dx2 = mt.sub(pt[0],pt2[0]);
		}
	}
	public void setTerminated(boolean t) {
		term = t;
	}
	public boolean getTerminated() {
		return term;
	}
	private void correctGradients(List<Circle<T>> cls,List<Float> iniGrad) {
		for(int i = 1;i<cls.size();i++) {
			Float[] ctr0 = mt.util.cast2FloatArray(cls.get(i-1).getCenter());
			Float[] ctr1 = mt.util.cast2FloatArray(cls.get(i-1).getCenter());
			if(ctr0[0]==ctr1[0]&&Math.abs(ctr0[1]-ctr1[1])==1)
				if(iniGrad.get(i-1)==-iniGrad.get(i))
					iniGrad.set(i,iniGrad.get(i)+0.01f);
		}
	}
	private void handleProgressListeners(int ac) {
		if(listeners!=null) {
			long t = System.currentTimeMillis();
			if(ac>INI_AC&&updateAc!=ac&&
					(ac%10==0||ac==accuracy||t-updateTime>=1000)) {
				updateTime = t;
				updateAc = ac;
				for(IProgressListener pl:listeners)
					pl.updateAccuracy(Math.min(ac,accuracy),accuracy);
			}
		}
	}
	private List<T[]> getReflPoints(List<Node> nodes) {
		ArrayList<T[]> reflPts = new ArrayList<T[]>();
		for(int i = 0;i<nodes.size();i++) {
			T[] pt = nodes.get(i).reflPnt;
			int digits = Math.max(mt.util.getScale(pt[0]),0)+accuracy;
			pt[0] = mt.util.copy(pt[0],digits);
			digits = Math.max(mt.util.getScale(pt[1]),0)+accuracy;
			pt[1] = mt.util.copy(pt[1],digits);
			reflPts.add(pt);
		}
		return reflPts;
	}
	private T[] calcMoveVector(Circle<T> c,T m,boolean dir) {
		T[] retVal = mt.util.array(2);
		T r10 = mt.mul(mt.TEN,c.getRadius());
		retVal[0] = r10;
		retVal[1] = mt.mul(r10,m);
		if(!dir)
			mt.neg(retVal);
		return retVal;
	}
	private void calcNewNode(Node nw,Node cur,Circle<T> c,int ac,T[] p1,T[] p2) {
		nw.reflPntDir = cur.reflPntDir;
		updateGradient(nw,cur);
		Profiler.getInstance().startTimeMeas("New Node");
		// To force convergence a damp factor is used here
		nw.m = dampGradientChange(nw.m,cur.m);
		T er = mt.abs(mt.sub(nw.m,cur.m));
		if(mt.cmp(er,mt.ZERO)==0)
			nw.ac = cur.ac;
		else
			nw.ac = -mt.util.getScale(er)-1;
		int errDigitPos = -mt.util.getErrorScale(nw.m);
		if(errDigitPos<ac+margin)
			nw.m = mt.util.copy(nw.m,mt.util.precision(nw.m)+ac-errDigitPos+margin);
		Profiler.getInstance().endTimeMeas("New Node");
		updatePoint(nw,c,p1,p2);
	}
	private T dampGradientChange(T m_new,T m_old) {
		T retVal = m_new;
		T er = mt.abs(mt.sub(m_new,m_old));
		if(mt.cmp(er,mt.ZERO)!=0) {
			int errPos = -mt.util.getScale(er);
			int firstPos = Math.max(-mt.util.getScale(m_new),-mt.util.getScale(m_old));
			boolean signChanged = mt.sign(m_new)!=mt.sign(m_old);
			if(errPos-firstPos<3||signChanged) {
				// Assumes that the angle between old and new reflection point is < pi/4.
				T s1 = mt.sqrt(mt.add(mt.mul(m_old,m_old),mt.ONE),accuracy+margin);
				T s2 = mt.sqrt(mt.add(mt.mul(m_new,m_new),mt.ONE),accuracy+margin);
				if(mt.cmp(mt.mul(m_new,m_old),mt.MINUS_ONE)<0) {
					T dx1 = mt.inv(s1);
					if(mt.sign(m_old)<0)
						dx1 = mt.neg(dx1);
					T dx2 = mt.inv(s2);
					if(mt.sign(m_new)<0)
						dx2 = mt.neg(dx2);
					T dx = mt.add(dx1,mt.mul(mt.sub(dx2,dx1),damp));
					retVal = mt.sqrt(mt.sub(mt.inv(mt.mul(dx,dx)),mt.ONE));
					if(mt.sign(dx)<0)
						retVal = mt.neg(retVal);
				} else {
					T dy1 = mt.div(m_old,s1);
					T dy2 = mt.div(m_new,s2);
					T dy = mt.add(dy1,mt.mul(mt.sub(dy2,dy1),damp));
					T dyy = mt.mul(dy,dy);
					retVal = m_old;
					if(mt.cmp(dyy,mt.ONE)<1) {
						T tmp = mt.sqrt(mt.sub(mt.ONE,dyy));
						if(mt.cmp(tmp,mt.ZERO)!=0)
							retVal = mt.div(dy,tmp);
					}
				}
				if(Globals.LOG_VERBOSE&&signChanged)
					Log.getIstc().logln("Sign changed: errPos= "+errPos+" retVal="+retVal);
			} else
				retVal = mt.add(m_old,mt.mul(mt.sub(m_new,m_old),damp));
		}
		return retVal;
	}
	private void updateGradient(Node nw,Node cur) {
		if(cur.ac<=INI_AC)
			nw.m = gradientExact(nw.dx1,nw.dy1,nw.dx2,nw.dy2,cur.m);
		else
			gradientApprox(nw,cur);
	}
	private T gradientExact(T dx1,T dy1,T dx2,T dy2,T m_old) {
		T tmp1 = mt.add(mt.mul(dy1,dx2),mt.mul(dy2,dx1));
		T tmp2 = mt.sub(mt.mul(dy1,dy2),mt.mul(dx1,dx2));
		T tmp3 = mt.ZERO;
		T tmp4 = mt.ZERO;
		if(mt.cmp(tmp1,mt.ZERO)!=0) {
			tmp1 = mt.div(tmp2,tmp1);
			tmp2 = mt.sqrt(mt.add(mt.mul(tmp1,tmp1),mt.ONE),accuracy+margin);
			tmp3 = mt.add(tmp1,tmp2);
			tmp4 = mt.sub(tmp1,tmp2);
		}
		// TODO Large gradients would cause a problem here
		double phiOld = Math.atan(m_old.doubleValue());
		double phi1 = Math.atan(tmp3.doubleValue());
		double phi2 = Math.atan(tmp4.doubleValue());
		T retVal = null;
		// The discrimination works only if the angle between old and new reflection
		// point is < pi/4. This assumption is only violated for a bad set of m_init's.
		if(linesAngleDelta(phi1,phiOld)<linesAngleDelta(phi2,phiOld))
			retVal = tmp3;
		else
			retVal = tmp4;
		return retVal;
	}
	private void gradientApprox(Node nw,Node cur) {
		if(nw.step1==-1||++nw.step1==15) {
			Profiler.getInstance().startTimeMeas("Newton Gradient Update");
			T a = mt.sub(mt.mul(nw.dy1,nw.dy2),mt.mul(nw.dx1,nw.dx2));
			T b = mt.add(mt.mul(nw.dy1,nw.dx2),mt.mul(nw.dy2,nw.dx1));
			GradientApproximator gradApprox = new GradientApproximator();
			gradApprox.a = a;
			gradApprox.b = b;
			try {
				nw.m = mt.approximateRoot(gradApprox.f,gradApprox.dfOverdx,cur.m);
			} catch(Exception e) {}
			boolean calcRef = nw.step1!=-1;
			if(!calcRef)
				calcRef = cur.ac+mt.util.getScale(a)>=10&&cur.ac+mt.util.getScale(b)>=10;
			if(calcRef) {
				nw.refAc1 = cur.ac;
				nw.refM1 = nw.m;
				nw.refDx1 = nw.dx1;
				nw.refDy1 = nw.dy1;
				nw.refDx2 = nw.dx2;
				nw.refDy2 = nw.dy2;
				T oneOverM = mt.inv(nw.m);
				nw.k1 = mt.div(mt.TWO,mt.mul(b,mt.add(mt.ONE,mt.mul(oneOverM,oneOverM))));
				nw.k2 = mt.mul(mt.ONE_OVER_TWO,mt.sub(nw.m,oneOverM));
				nw.step1 = 0;
			}
			Profiler.getInstance().endTimeMeas("Newton Gradient Update");
		} else {
			Profiler.getInstance().startTimeMeas("Linear Gradient Update");
			T ddx1 = mt.sub(nw.dx1,nw.refDx1);
			T ddy1 = mt.sub(nw.dy1,nw.refDy1);
			T ddx2 = mt.sub(nw.dx2,nw.refDx2);
			T ddy2 = mt.sub(nw.dy2,nw.refDy2);
			T da1 = mt.add(mt.mul(ddx1,nw.refDx2),mt.mul(ddx2,nw.refDx1));
			T da = mt.sub(mt.add(mt.mul(ddy1,nw.refDy2),mt.mul(ddy2,nw.refDy1)),da1);
			T db1 = mt.add(mt.mul(ddx1,nw.refDy2),mt.mul(ddy2,nw.refDx1));
			T db = mt.add(mt.add(mt.mul(ddy1,nw.refDx2),mt.mul(ddx2,nw.refDy1)),db1);
			T dm = mt.mul(nw.k1,mt.sub(da,mt.mul(nw.k2,db)));
			nw.m = mt.add(nw.refM1,dm);
			Profiler.getInstance().endTimeMeas("Linear Gradient Update");
		}
	}
	public void setLog(boolean l) {
		this.log = l;
	}
	public int getAccuracy() {
		return accuracy;
	}
	public void setAccuracy(int ac) {
		this.accuracy = ac;
	}
	public int getNodeAccuracy(int pos) {
		return curNodes.get(pos).ac;
	}
	public void setOptimalDamp(T od) {
		this.opt_damp = mt.util.copy(od,margin);
	}
	public T getOptimalDamp() {
		return opt_damp;
	}
	public void setInitialDamp(T d) {
		this.damp = mt.util.copy(d,margin);
	}
	public void addProgressListener(IProgressListener l) {
		if(listeners==null)
			listeners = new ArrayList<IProgressListener>();
		if(l!=null)
			listeners.add(l);
	}
	private void updatePoint(Node rn,Circle<T> c,T[] p1,T[] p2) {
		T relX;
		T relY;
		int digits = rn.ac+mt.util.getScale(rn.m);
		boolean b = digits>10;
		if(!b||rn.step2==-1||++rn.step2==20) {
			Profiler.getInstance().startTimeMeas("Point Update");
			T tmp1;
			T tmp2;
			int p = mt.util.precision(rn.m);
			T m = p<=Util.MAX_INT-5&&b?mt.util.copy(rn.m,p+5):rn.m;
			if(rn.sqrtOneOverM2PlusOne!=null) {
				T tmp = mt.util.copy(rn.sqrtOneOverM2PlusOne,mt.util.precision(m));
				tmp2 = mt.add(tmp,mt.inv(mt.mul(tmp,mt.add(mt.mul(m,m),mt.ONE))));
				tmp2 = mt.mul(mt.ONE_OVER_TWO,tmp2);
				tmp1 = mt.mul(tmp2,tmp2);
			} else {
				tmp1 = mt.inv(mt.add(mt.mul(m,m),mt.ONE));
				tmp2 = mt.sqrt(tmp1,accuracy+margin);
			}
			relX = mt.mul(c.getRadius(),tmp2);
			relY = mt.mul(m,relX);
			if(b) {
				rn.step2 = 0;
				rn.refAc2 = rn.ac;
				rn.refRelX = relX;
				rn.refRelY = relY;
				rn.refRelXOverM2PlusOne = mt.mul(rn.refRelX,tmp1);
				rn.refRelYOverM2PlusOne = mt.mul(rn.refRelY,tmp1);
				rn.refM2 = m;
				if(digits>100)
					rn.sqrtOneOverM2PlusOne = tmp2;
			} else
				rn.step2 = -1;
			Profiler.getInstance().endTimeMeas("Point Update");
		} else {
			Profiler.getInstance().startTimeMeas("Linear Point Update");
			T dm = mt.sub(rn.m,rn.refM2);
			relX = mt.sub(rn.refRelX,mt.mul(rn.refRelYOverM2PlusOne,dm));
			relY = mt.add(rn.refRelY,mt.mul(rn.refRelXOverM2PlusOne,dm));
			Profiler.getInstance().endTimeMeas("Linear Point Update");
		}
		rn.reflPnt = calcReflPoint(c.getCenter(),p1,p2,relX,relY);
		return;
	}
	private T[] calcReflPoint(T[] ctr,T[] p1,T[] p2,T dx,T dy) {
		Profiler.getInstance().startTimeMeas("Calc Point");
		float ctrXf = ctr[0].floatValue();
		float ctrYf = ctr[1].floatValue();
		float dxf = dx.floatValue();
		float dyf = dy.floatValue();
		float pt1Xf = ctrXf+dxf;
		float pt1Yf = ctrYf+dyf;
		float pt2Xf = ctrXf-dxf;
		float pt2Yf = ctrYf-dyf;
		float p1x = p1[0].floatValue();
		float p1y = p1[1].floatValue();
		float p2x = p2[0].floatValue();
		float p2y = p2[1].floatValue();
		float d1 = distSqrd(pt1Xf,pt1Yf,p1x,p1y,p2x,p2y);
		float d2 = distSqrd(pt2Xf,pt2Yf,p1x,p1y,p2x,p2y);
		T[] retVal = mt.util.array(2);
		if(d2>=d1) {
			retVal[0] = mt.add(ctr[0],dx);
			retVal[1] = mt.add(ctr[1],dy);
		} else {
			retVal[0] = mt.sub(ctr[0],dx);
			retVal[1] = mt.sub(ctr[1],dy);
		}
		Profiler.getInstance().endTimeMeas("Calc Point");
		return retVal;
	}
	private static float distSqrd(float px,float py,float p1x,float p1y,float p2x,float p2y) {
		float dx1 = px-p1x;
		float dy1 = py-p1y;
		float dx2 = px-p2x;
		float dy2 = py-p2y;
		return dx1*dx1+dy1*dy1+dx2*dx2+dy2*dy2;
	}
	/**
	 * Returns the smaller of the two intersection angles between two lines.
	 * The method arguments are the gradient angles of the lines. The return
	 * value is therfore <= PI/2
	 */
	private static double linesAngleDelta(double a1,double a2) {
		double delta;
		if(a1>=0==a2>=0)
			delta = a2>a1?a2-a1:a1-a2;
		else {
			double sum = (a1>=0?a1:-a1)+(a2>=0?a2:-a2);
			double complSum = 2*Math.PI-sum;
			delta = sum>complSum?complSum:sum;
		}
		double complDelta = Math.PI-delta;
		return delta>complDelta?complDelta:delta;
	}
	public interface IProgressListener {
		void updateAccuracy(int digits,int maxDigits);
	}
	private void printInfo(int steps,long t0,List<Node> nodes,byte offset) {
		if(log) {
			printReflectionNodes(nodes,offset);
			int t = (int)((System.currentTimeMillis()-t0)/1000f);
			Log.getIstc().log(", Time = "+t+" sec");
			if(nodes.size()>0) {
				float speed = ((int)(accuracy*100f/steps+0.5f))/100f;
				Log.getIstc().log(", Steps = "+steps);
				Log.getIstc().log("\nDigits/Steps = "+speed);
				Log.getIstc().log(", Damp = "+damp);
				Log.getIstc().log(", Margin = "+margin+" digits");
			}
			Log.getIstc().log("\n");
		}
	}
	private void printReflectionNodes(List<Node> nodes,byte offset) {
		if(log&&nodes.size()>0) {
			Log.getIstc().logln("**********************************************************");
			for(String h:HEADERS)
				Log.getIstc().log(h);
			Log.getIstc().log("\n");
			float dphiMax = Float.MIN_VALUE;
			for(int i = 0;i<nodes.size();i++) {
				Node rn = nodes.get(i);
				Log.getIstc().logf(i+1+offset+"",HEADERS[0].length());
				String delta = "";
				if(mt.cmp(rn.m0,rn.m)!=0) {
					Log.getIstc().logf("1e-"+rn.ac+"",HEADERS[1].length());
					float phi = (float)Math.atan(rn.m.floatValue());
					float phi0 = (float)Math.atan(rn.m0.floatValue());
					float dphi = (float)(linesAngleDelta(phi,phi0)*180/Math.PI);
					dphi = (int)(dphi*10+0.5)/10f;
					delta = dphi+"°"; 
					dphiMax = dphiMax<dphi?dphi:dphiMax;
				} else
					Log.getIstc().logf("",HEADERS[1].length());
				String m = mt.util.toPlainString(rn.m);
				m = m.substring(0,Math.min(32,m.length()));
				Log.getIstc().logf(m,HEADERS[2].length());
				Log.getIstc().logf(delta,HEADERS[3].length());
				if(Globals.LOG_VERBOSE) {
					if(mt.cmp(rn.dx1,mt.ZERO)==0)
						Log.getIstc().logf("infinity",HEADERS[2].length());
					else {
						T m1 = mt.util.copy(mt.div(rn.dy1,rn.dx1),20);
						Log.getIstc().logf(""+m1,HEADERS[2].length());
					}
					if(mt.cmp(rn.dx2,mt.ZERO)==0)
						Log.getIstc().logf("infinity",HEADERS[2].length());
					else {
						T m2 = mt.util.copy(mt.div(rn.dy2,rn.dx2),20);
						Log.getIstc().logf(""+m2,HEADERS[2].length());
					}
					Log.getIstc().logf(""+mt.util.copy(rn.reflPnt[0],10),17);
					Log.getIstc().logf(""+mt.util.copy(rn.reflPnt[1],10),17);
				}
				Log.getIstc().log("\n");
			}
			Log.getIstc().logln("**********************************************************");
			if(dphiMax!=Float.MIN_VALUE)
				Log.getIstc().log("Max. Delta = "+dphiMax+"°");
		}
	}
	private final class GradientApproximator {

		private T a;
		private T b;
		private T tmp2;
		private T tmp3;
		private final IFunction<T> f;
		private final IFunction<T> dfOverdx;
		private GradientApproximator() {
			f = new IFunction<T>() {

				@Override
				public T f(T x) {
					tmp2 = mt.mul(mt.TWO,a);
					tmp3 = mt.mul(b,x);
					return mt.sub(mt.mul(tmp3,x),mt.add(b,mt.mul(tmp2,x)));
				}
			};
			dfOverdx = new IFunction<T>() {

				@Override
				public T f(T x) {
					return mt.sub(mt.mul(tmp3,mt.TWO),tmp2);
				}
			};
		}
	}
	private final class Node {

		private T[] reflPnt;
		private T m0;
		private T m;
		private T dy1;
		private T dx1;
		private T dy2;
		private T dx2;
		// The accuracy of the node in digits
		private int ac = 0;
		private boolean reflPntDir;
		// Temporaries for fast dm calculation
		private int step1 = -1;
		private int refAc1;
		private T refM1;
		private T refDy1;
		private T refDx1;
		private T refDy2;
		private T refDx2;
		private T k1;
		private T k2;
		// Temporaries for linearized reflection point calculation
		private int step2 = -1;
		private int refAc2;
		private T refM2;
		private T refRelX;
		private T refRelY;
		private T refRelXOverM2PlusOne;
		private T refRelYOverM2PlusOne;
		private T sqrtOneOverM2PlusOne;
		Node(LightPatternSolver<T>.Node rn) {
			this.m0 = rn.m0;
			this.refAc1 = rn.refAc1;
			this.step1 = rn.step1;
			this.refM1 = rn.refM1;
			this.refDy1 = rn.refDy1;
			this.refDx1 = rn.refDx1;
			this.refDy2 = rn.refDy2;
			this.refDx2 = rn.refDx2;
			this.k1 = rn.k1;
			this.k2 = rn.k2;
			this.step2 = rn.step2;
			this.refAc2 = rn.refAc2;
			this.refM2 = rn.refM2;
			this.refRelX = rn.refRelX;
			this.refRelY = rn.refRelY;
			this.refRelXOverM2PlusOne = rn.refRelXOverM2PlusOne;
			this.refRelYOverM2PlusOne = rn.refRelYOverM2PlusOne;
			this.sqrtOneOverM2PlusOne = rn.sqrtOneOverM2PlusOne;
		}
		Node() {}
	}
}
